#
#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Lead Developers: Dan Lovell and Jay Baxter
#   Authors: Dan Lovell, Baxter Eaves, Jay Baxter, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import argparse
import sys
from collections import Counter
#
import numpy
import pylab
pylab.ion()
pylab.show()
#
import crosscat.utils.file_utils as fu
import crosscat.utils.sample_utils as su


# parse some arguments
parser = argparse.ArgumentParser()
parser.add_argument('pkl_name', type=str)
parser.add_argument('--inf_seed', default=0, type=int)
args = parser.parse_args(['/usr/local/crosscat/cython_code/iter_90_pickled_state.pkl.gz'])
pkl_name = args.pkl_name
inf_seed = args.inf_seed

random_state = numpy.random.RandomState(inf_seed)
# FIXME: getting weird error on conversion to int: too large from inside pyx
def get_next_seed(max_val=32767): # sys.maxint):
    return random_state.randint(max_val)

# resume from saved name
save_dict = fu.unpickle(pkl_name)
M_c = save_dict['M_c']
X_L = save_dict['X_L']
X_D = save_dict['X_D']
T = save_dict['T']
num_cols = len(X_L['column_partition']['assignments'])
row_idx = 205
col_idx = 13
Q = [(row_idx, col_idx)]
imputed, confidence = su.impute_and_confidence(
    M_c, X_L, X_D, Y=None, Q=Q, n=400, get_next_seed=get_next_seed)

T_array = numpy.array(T)
which_view_idx = X_L['column_partition']['assignments'][col_idx]
X_D_i = numpy.array(X_D[which_view_idx])
which_cluster_idx = X_D_i[row_idx]
which_rows_match_indices = numpy.nonzero(X_D_i==which_cluster_idx)[0]
cluster_vals = T_array[which_rows_match_indices, col_idx]
all_vals = T_array[:, col_idx]
cluster_counter = Counter(cluster_vals)
cluster_ratio = float(cluster_counter[imputed]) / sum(cluster_counter.values())
all_counter = Counter(all_vals)
all_ratio = float(all_counter[imputed]) / sum(all_counter.values())
print()
print('imputed: %s' % imputed)
print('all_ratio: %s' % all_ratio)
print('cluster_ratio: %s' % cluster_ratio)
print('confidence: %s' % confidence)



#
#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Lead Developers: Dan Lovell and Jay Baxter
#   Authors: Dan Lovell, Baxter Eaves, Jay Baxter, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import functools
#
from IPython.parallel import Client
#
import crosscat
import crosscat.LocalEngine as LE


def partialize(func, args_dict, dview):
    # why is this push necessary?
    dview.push(args_dict, block=True)
    helper = functools.partial(func, **args_dict)
    return helper


class IPClusterEngine(LE.LocalEngine):
    """A simple interface to the Cython-wrapped C++ engine

    IPClusterEngine

    """

    def __init__(self, config_filename=None, profile=None, seed=None, sshkey=None, packer='json'):
        """Initialize a IPClusterEngine

        Do IPython.parallel operations to set up cluster and generate mapper.

        """
        super(IPClusterEngine, self).__init__(seed=seed)
        rc = Client(config_filename, profile=profile, sshkey=sshkey, packer=packer)
        # FIXME: add a warning if environment in direct view is not 'empty'?
        #        else, might become dependent on an object created in
        #        environemnt in a prior run
        dview = rc.direct_view()
        lview = rc.load_balanced_view()
        with dview.sync_imports(local=True):
            import crosscat
        mapper = lambda f, tuples: self.lview.map(f, *tuples)
        # if you're trying to debug issues, consider clearning to start fresh
        # rc.clear(block=True)
        #
        self.rc = rc
        self.dview = dview
        self.lview = lview
        self.mapper = mapper
        self.do_initialize = None
        self.do_analyze = None
        return

    def get_initialize_arg_tuples(self, M_c, M_r, T, initialization,
            row_initialization, n_chains):
        args_dict = dict(M_c=M_c, M_r=M_r, T=T, initialization=initialization,
                row_initialization=row_initialization)
        do_initialize = partialize(crosscat.LocalEngine._do_initialize,
                args_dict, self.dview)
        seeds = [self.get_next_seed() for seed_idx in range(n_chains)]
        arg_tuples = [seeds]
        #
        self.do_initialize = do_initialize
        return arg_tuples

    def get_analyze_arg_tuples(self, M_c, T, X_L, X_D, kernel_list=(), n_steps=1, c=(), r=(),
                max_iterations=-1, max_time=-1, diagnostic_func_dict=None, every_N=1):
        n_chains = len(X_L)
        args_dict = dict(M_c=M_c, T=T, kernel_list=kernel_list, n_steps=n_steps,
                c=c, r=r, max_iterations=max_iterations, max_time=max_time,
                diagnostic_func_dict=diagnostic_func_dict, every_N=every_N)
        do_analyze = partialize(crosscat.LocalEngine._do_analyze_with_diagnostic,
                args_dict, self.dview)
        seeds = [self.get_next_seed() for seed_idx in range(n_chains)]
        arg_tuples = [seeds, X_L, X_D]
        #
        self.do_analyze = do_analyze
        return arg_tuples



import crosscat.cython_code.MultinomialComponentModel as mcm
import math
import random
import numpy
import six

from scipy.special import gammaln as gammaln

from crosscat.utils.general_utils import logmeanexp

next_seed = lambda : random.randrange(2147483647)

###############################################################################
#   Input-checking and exception-handling functions
###############################################################################
def check_type_force_float(x, name):
    """
    If an int is passed, convert it to a float. If some other type is passed,
    raise an exception.
    """
    if type(x) is int:
        return float(x)
    elif type(x) is not float and type(x) is not numpy.float64:
        raise TypeError("%r should be a float" % (name,))
    else:
        return x

def counts_to_data(counts):
    """
    Converts a vector of counts to data.
    """
    assert type(counts) is list or type(counts) is numpy.ndarray
    K = len(counts)
    N = int(sum(counts))
    X = []
    for k in range(K):
        i = 0
        while i < counts[k]:
            X.append([k])
            i += 1

        assert i == counts[k]

    assert len(X) == N

    random.shuffle(X)
    X = numpy.array(X, dtype=float)

    return X


def check_data_type_column_data(X):
    """
    Makes sure that X is a numpy array and that it is a column vector
    """
    if type(X) is list:
        X = numpy.array(X)

    if type(X) is not numpy.ndarray:
        raise TypeError("X should be type numpy.ndarray or a list")

    if len(X.shape) == 2 and X.shape[1] > 1:
        raise TypeError("X should have a single column.")


def check_model_parameters_dict(model_parameters_dict):

    if type(model_parameters_dict) is not dict:
        raise TypeError("model_parameters_dict should be a dict")

    keys = ['weights']

    for key in keys:
        if key not in model_parameters_dict:
            raise KeyError("model_parameters_dict should have key %r" % (key,))

    for key, value in six.iteritems(model_parameters_dict):
        if key == "weights":
            if type(value) is not list:
                raise TypeError("model parameters dict key 'weights' should be a list")
            if type(value[0]) is list:
                raise TypeError("weights should not be a list of lists, should be a list of floats")
            if math.fabs(sum(value) - 1.0) > .00000001:
                raise ValueError("model parameters dict key 'weights' should sum to 1.0")
        else:
            raise KeyError("invalid key, %r, for model parameters dict" % (key,))

def check_hyperparameters_dict(hyperparameters_dict):

    # 'fixed' key is not necessary for user-defined hyperparameters
    keys = ['dirichlet_alpha', 'K']

    for key in keys:
        if key not in hyperparameters_dict:
            raise KeyError("hyperparameters_dict should have key %r" % (key,))

    for key, value in six.iteritems(hyperparameters_dict):
        if key == "K":
            if type(value) is not int:
                raise TypeError("hyperparameters dict entry K should be an int")

            if value < 1:
                raise ValueError("hyperparameters dict entry K should be greater than 0")
        elif key == "dirichlet_alpha":
            if not isinstance(value, (float, numpy.float64, int)):
                raise TypeError("hyperparameters dict entry dirichlet_alpha should be a float or int")
            if value <= 0.0:
                raise ValueError("hyperparameters dict entry dirichlet_alpha should be greater than 0")
        elif key == "fixed":
            pass
        else:
            raise KeyError("invalid key, %r, for hyperparameters dict" % (key,))

def check_data_vs_k(X,K):
    if type(X) is numpy.ndarray:
        X = X.flatten(1)
        X = X.tolist()
    K_data = len(set(X))
    if K_data > K:
        raise ValueError("the number of items in the data is greater than K")

###############################################################################
#   The class extension
###############################################################################
class p_MultinomialComponentModel(mcm.p_MultinomialComponentModel):

    model_type = 'symmetric_dirichlet_discrete'
    cctype = 'multinomial'

    @classmethod
    def from_parameters(cls, N, params=None, hypers=None, gen_seed=0):
        """
        Initialize a continuous component model with sufficient statistics
        generated from random data.
        Inputs:
          N: the number of data points
          params: a dict with the following keys
              weights: a K-length list that sums to 1.0
          hypers: a dict with the following keys
              K: the number of categories
              dirichlet_alpha: Dirichlet alpha parameter. The distribution is
              symmetric so only one value is needed
          gen_seed: an integer from which the rng is seeded
        """

        if type(N) is not int:
            raise TypeError("N should be an int")
        if type(gen_seed) is not int:
            raise TypeError("gen_seed should be an int")

        # if the parameters dict or the hypers dict exist, validate them
        if params is not None:
            check_model_parameters_dict(params)

        if hypers is not None:
            check_hyperparameters_dict(hypers)

        random.seed(gen_seed)
        numpy.random.seed(gen_seed)

        # get the number of categories
        if params is None:
            if hypers is None:
                K = int(N/2.0)
            else:
                K = int(hypers['K'])
            weights = numpy.random.random((1,K))
            weights = weights/numpy.sum(weights)
            weights = weights.tolist()[0]
            assert len(weights) == K
            params = dict(weights=weights)

            check_model_parameters_dict(params)
        else:
            K = len(params['weights'])
            if hypers:
                if K != hypers['K']:
                    raise ValueError("K in params does not match K in hypers")

        # generate synthetic data
        counts = numpy.array(numpy.random.multinomial(N, params['weights']), dtype=int)

        X = counts_to_data(counts)

        check_data_type_column_data(X)

        # generate the sufficient statistics
        suffstats = dict()
        for k in range(K):
            suffstats[str(k)] = counts[k]

        if hypers is None:
            hypers = cls.draw_hyperparameters(X, n_draws=1, gen_seed=next_seed())[0]
            check_hyperparameters_dict(hypers)

        # hypers['K'] = check_type_force_float(hypers['K'], "hypers['K']")
        hypers['dirichlet_alpha'] = check_type_force_float(hypers['dirichlet_alpha'], "hypers['dirichlet_alpha']")

        # add fixed parameter to hyperparameters
        hypers['fixed'] = 0.0

        suffstats = {'counts':suffstats}

        return cls(hypers, float(N), **suffstats)

    @classmethod
    def from_data(cls, X, hypers=None, gen_seed=0):
        """
        Initialize a continuous component model with sufficient statistics
        generated from data X
        Inputs:
            X: a column of data (numpy)
            hypers: a dict with the following keys
              K: the number of categories
              dirichlet_alpha: Dirichlet alpha parameter. The distribution is
              symmetric so only one value is needed
            gen_seed: a int to seed the rng
        """
        # FIXME: Figure out a way to accept a list of strings
        check_data_type_column_data(X)
        if type(gen_seed) is not int:
            raise TypeError("gen_seed should be an int")

        random.seed(gen_seed)
        numpy.random.seed(gen_seed)

        if hypers is None:
            hypers = cls.draw_hyperparameters(X, gen_seed=next_seed())[0]
            check_hyperparameters_dict(hypers)
        else:
            check_hyperparameters_dict(hypers)
            K = hypers['K']
            check_data_vs_k(X,K)

        hypers['dirichlet_alpha'] = check_type_force_float(hypers['dirichlet_alpha'], "hypers['dirichlet_alpha']")

        N = len(X)
        K = hypers['K']

        counts = [0]*K
        for x in X:
            try:
                counts[int(x)] += 1
            except IndexError:
                raise IndexError

        # generate the sufficient statistics
        suffstats = dict()
        for k in range(int(K)):
            suffstats[str(k)] = counts[k]

        suffstats = {'counts':suffstats}

        hypers['fixed'] = 0.0

        return cls(hypers, float(N), **suffstats)

    def sample_parameters_given_hyper(self, gen_seed=0):
        """
        Samples weights given the current hyperparameters
        Inputs:
            gen_seed: integer used to seed the rng
        """
        if type(gen_seed) is not int:
            raise TypeError("gen_seed should be an int")

        random.seed(gen_seed)
        numpy.random.seed(gen_seed)

        hypers = self.get_hypers()
        dirichlet_alpha = hypers[b'dirichlet_alpha']
        K = hypers[b'K']

        alpha = numpy.array([dirichlet_alpha]*int(K))

        weights = numpy.random.dirichlet(alpha)
        weights = weights.tolist()

        params = {'weights': weights}

        return params

    def uncollapsed_likelihood(self, X, params):
        """
        Calculates the score of the data X under this component model with mean
        mu and precision rho.
        Inputs:
            X: A column of data (numpy)
            params: a dict with the following keys
                weights: a list of category weights (should sum to 1)
        """
        check_data_type_column_data(X)
        check_model_parameters_dict(params)

        hypers = self.get_hypers()

        assert len(params['weights']) == int(hypers[b'K'])

        dirichlet_alpha = hypers[b'dirichlet_alpha']
        K = float(hypers[b'K'])
        check_data_vs_k(X,K)

        weights = numpy.array(params['weights'])

        log_likelihood = self.log_likelihood(X, params)
        logB = gammaln(dirichlet_alpha)*K - gammaln(dirichlet_alpha*K)
        log_prior = -logB + numpy.sum((dirichlet_alpha-1.0)*numpy.log(weights))

        log_p = log_likelihood + log_prior

        return log_p

    @staticmethod
    def log_likelihood(X, params):
        """
        Calculates the log likelihood of the data X given mean mu and precision
        rho.
        Inputs:
            X: a column of data (numpy)
            params: a dict with the following keys
                weights: a list of categories weights (should sum to 1)
        """
        check_data_type_column_data(X)
        check_model_parameters_dict(params)

        N = len(X)
        K = len(params['weights'])
        check_data_vs_k(X,K)
        counts= numpy.bincount(X,minlength=K)

        weights = numpy.array(params['weights'])

        A = gammaln(N+1)-numpy.sum(gammaln(counts+1))
        B = numpy.sum(counts*numpy.log(weights));

        log_likelihood = A+B

        return log_likelihood

    @staticmethod
    def log_pdf(X, params):
        """
        Calculates the log pdf of the data X given mean mu and precision
        rho.
        Inputs:
            X: a column of data (numpy)
            params: a dict with the following keys
                weights: a list of categories weights (should sum to 1)
        """
        check_data_type_column_data(X)
        check_model_parameters_dict(params)

        N = len(X)

        weights = numpy.array(params['weights'])

        lpdf = []
        for x in X:
            w = weights[int(x)]
            if w == 0.0 or w == 0:
                lpdf.append(float('-Inf'))
            else:
                lpdf.append(math.log(w))

        return numpy.array(lpdf)

    def brute_force_marginal_likelihood(self, X, n_samples=10000, gen_seed=0):
        """
        Calculates the log marginal likelihood via brute force method in which
        parameters (weights) are repeatedly drawn from the prior, the
        likelihood is calculated for each set of parameters, then the average is
        taken.
        Inputs:
            X: A column of data (numpy)
            n_samples: the number of draws
            gen_Seed: seed for the rng
        """
        check_data_type_column_data(X)

        if type(n_samples) is not int:
            raise TypeError("n_samples should be an int")
        if n_samples <= 0:
            raise ValueError("n_samples should be greater than 0")
        if type(gen_seed) is not int:
            raise TypeError("gen_seed should be an int")

        hypers = self.get_hypers()
        K = hypers[b'K']
        check_data_vs_k(X,K)

        random.seed(gen_seed)
        log_likelihoods = [0]*n_samples
        for i in range(n_samples):
            params = self.sample_parameters_given_hyper(gen_seed=next_seed())
            log_likelihoods[i] = self.log_likelihood(X, params)

        log_marginal_likelihood = logmeanexp(log_likelihoods)

        return log_marginal_likelihood

    @staticmethod
    def generate_discrete_support(params):
        """
        Returns the a sequential list of the number of categories
        Inputs:
            params: a dict with entries 'mu' and 'rho'
        """
        check_model_parameters_dict(params)

        return list(range(len(params['weights'])))


    @staticmethod
    def draw_hyperparameters(X, n_draws=1, gen_seed=0):
        """
        Draws hyperparameters dirichlet_alpha from the same distribution that
        generates the grid in the C++ code.
        Inputs:
             X: a column of data or an int which acts as K. If a data array is
                provided, K is assumed to be max(X)+1
             n_draws: the number of draws
             gen_seed: seed the rng
        Output:
            A list of dicts of draws where each entry has keys 'dirichlet_alpha'
            and 'K'. K is defined by the data and will be the same for each samples
        """
        if type(X) is list or type(X) is numpy.ndarray:
            check_data_type_column_data(X)
            K = int(max(X)+1)
        elif type(X) is int:
            if X < 1:
                raise ValueError("If X is an int, it should be greatert than 1")
            K = X
        else:
            raise TypeError("X should be an array or int")

        if type(n_draws) is not int:
            raise TypeError("n_draws should be an int")

        if type(gen_seed) is not int:
            raise TypeError("gen_seed should be an int")

        random.seed(gen_seed)

        samples = []

        # get draw ranges
        alpha_draw_range = (0.1, math.log(K))


        for i in range(n_draws):
            alpha = math.exp(random.uniform(alpha_draw_range[0], alpha_draw_range[1]))

            this_draw = dict(dirichlet_alpha=alpha, K=K)

            samples.append(this_draw)

        assert len(samples) == n_draws

        return samples

    @staticmethod
    def generate_data_from_parameters(params, N, gen_seed=0):
        """
        returns a set of intervals over which the component model pdf is
        supported.
        Inputs:
            params: a dict with entries 'weights'
            N: number of data points
        """
        if type(N) is not int:
            raise TypeError("N should be an int")

        if N <= 0:
            raise ValueError("N should be greater than 0")

        if type(params) is not dict:
            raise TypeError("params should be a dict")

        check_model_parameters_dict(params)

        # multinomial draw
        counts = numpy.array(numpy.random.multinomial(N, params['weights']), dtype=int)

        X = counts_to_data(counts)

        assert len(X) == N

        return X



#
#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Lead Developers: Dan Lovell and Jay Baxter
#   Authors: Dan Lovell, Baxter Eaves, Jay Baxter, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import crosscat.utils.general_utils as gu


class EngineTemplate(object):

    def __init__(self, seed=None):
        self.seed_generator = gu.int_generator(seed)

    def get_next_seed(self):
        return self.seed_generator.next()

    def initialize(self, M_c, M_r, T, initialization='from_the_prior'):
        M_c, M_r, X_L, X_D = dict(), dict(), dict(), []
        return X_L, X_D

    def analyze(self, M_c, T, X_L, X_D, kernel_list=(), n_steps=1, c=(), r=(),
                max_iterations=-1, max_time=-1, do_diagnostics=False,
                diagnostics_every_N=1,
                ROW_CRP_ALPHA_GRID=(), COLUMN_CRP_ALPHA_GRID=(),
                S_GRID=(), MU_GRID=(),
                N_GRID=31,
                ):
        X_L_prime, X_D_prime = dict(), []
        return X_L_prime, X_D_prime

    def simple_predictive_sample(self, M_c, X_L, X_D, Y, Q, n=1):
        samples = []
        return samples

    def simple_predictive_probability(self, M_c, X_L, X_D, Y, Q, n):
        p = None
        return p

    def simple_predictive_probability_multistate(self, M_c, X_L_list, X_D_list, Y, Q, n):
        p = None
        return p

    def mutual_information(self, M_c, X_L_list, X_D_list, Q, n_samples=1000):
        return None

    def row_structural_typicality(self, X_L_list, X_D_list, row_id):
        return None

    def column_structural_typicality(self, X_L_list, col_id):
        return None

    def predictive_probability(self, M_c, X_L, X_D, T, Q, n=1):
        return None

    def predictive_probability_multistate(self, M_c, X_L_list, X_D_list, T, Q, n=1):
        return None

    def similarity(self, M_c, X_L_list, X_D_list, given_row_id, target_row_id, target_columns=None):
        return None

    def impute(self, M_c, X_L, X_D, Y, Q, n):
        e = None
        return e

    def impute_and_confidence(self, M_c, X_L, X_D, Y, Q, n):
        e, confidence = None, None
        return e, confidence

    def conditional_entropy(M_c, X_L, X_D, d_given, d_target,
                            n=None, max_time=None):
        e = None
        return e

    def predictively_related(self, M_c, X_L, X_D, d,
                                           n=None, max_time=None):
        m = []
        return m

    def contextual_structural_similarity(self, X_D, r, d):
        s = []
        return s

    def structural_similarity(self, X_D, r):
        s = []
        return s

    def structural_anomalousness_columns(self, X_D):
        a = []
        return a

    def structural_anomalousness_rows(self, X_D):
        a = []
        return a

    def predictive_anomalousness(self, M_c, X_L, X_D, T, q, n):
        a = []
        return a




#
#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Lead Developers: Dan Lovell and Jay Baxter
#   Authors: Dan Lovell, Baxter Eaves, Jay Baxter, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import print_function
import argparse
#
import crosscat.utils.api_utils as au
import crosscat.utils.file_utils as fu


# parse some arguments
parser = argparse.ArgumentParser()
parser.add_argument('pkl_name', type=str)
parser.add_argument('--hostname', default='localhost', type=str)
parser.add_argument('--seed', default=0, type=int)
parser.add_argument('--start_id', default=0, type=int)
args = parser.parse_args()
pkl_name = args.pkl_name
hostname = args.hostname
seed = args.seed
id = args.start_id

# settings
URI = 'http://' + hostname + ':8007'
print('URI: ', URI)

save_dict = fu.unpickle(pkl_name)
method_name = 'analyze'
args_dict = dict()
args_dict['M_c'] = save_dict['M_c']
args_dict['T'] = save_dict['T']
args_dict['X_L'] = save_dict['X_L']
args_dict['X_D'] = save_dict['X_D']
args_dict['kernel_list'] = 'kernel_list'
args_dict['n_steps'] = 1
args_dict['c'] = 'c'
args_dict['r'] = 'r'
args_dict['max_iterations'] = 'max_iterations'
args_dict['max_time'] = 'max_time'
out, id = au.call(method_name, args_dict, URI, id)
X_L_prime, X_D_prime = out



#
#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Lead Developers: Dan Lovell and Jay Baxter
#   Authors: Dan Lovell, Baxter Eaves, Jay Baxter, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import print_function
import itertools
import csv
import os

ari_filename = 'ari_convergence_results.csv'
n_steps = 500
block_size = 20
run_script = False
#
base_str = ' '.join([
  'python convergence_test.py',
  '--max_rows %s',
  '--num_cols %s',
  '--num_clusters %s',
  '--num_views %s',
  '--block_size %s' % block_size,
  '--num_transitions %s' % n_steps,
  '--ari_logfile %s' % ari_filename,
  '>>out 2>>err',
  ])

# num_rows_list = [100, 400, 1000, 4000, 10000]
# num_cols_list = [4, 8, 16, 24, 32]
# num_clusters_list = [10, 20, 30, 40, 50]
# num_splits_list = [1, 2, 3, 4, 5]

num_rows_list = [200, 400]
num_cols_list = [4, 8]
num_clusters_list = [5, 10]
num_splits_list = [2,4]

# First create the headers in the output file - the convergence test script does not write headers
with open(ari_filename, 'w') as outfile:
  csvwriter=csv.writer(outfile,delimiter=',')
  header = ['Time', 'num_transitions', 'block_size', 'num_rows','num_cols','num_views','num_clusters','ari_views','ari_table']
  csvwriter.writerow(header)

outfile.close()
count = 0
script_name = 'convergence_testing_script.sh'
with open(script_name, 'w') as script_file:
    take_product_of = [num_rows_list, num_cols_list, num_clusters_list, num_splits_list]
    for num_rows, num_cols, num_clusters, num_splits \
        in itertools.product(*take_product_of):
        this_base_str = base_str % (num_rows, num_cols, num_clusters, num_splits)
        print(this_base_str)
        count = count + 1
        script_file.write(this_base_str + '\n')

script_file.close()

if run_script:
  os.system('bash convergence_testing_script.sh')



import numpy
#
import crosscat.utils.convergence_test_utils


def get_logscore(p_State):
    return p_State.get_marginal_logp()

def get_num_views(p_State):
    return len(p_State.get_X_D())

def get_column_crp_alpha(p_State):
    return p_State.get_column_crp_alpha()

def get_ari(p_State):
    # requires environment: {view_assignment_truth}
    # requires import: {crosscat.utils.convergence_test_utils}
    X_L = p_State.get_X_L()
    ctu = crosscat.utils.convergence_test_utils
    return ctu.get_column_ARI(X_L, view_assignment_truth)

def get_mean_test_ll(p_State):
    # requires environment {M_c, T, T_test}
    # requires import: {crosscat.utils.convergence_test_utils}
    X_L = p_State.get_X_L()
    X_D = p_State.get_X_D()
    ctu = crosscat.utils.convergence_test_utils
    return ctu.calc_mean_test_log_likelihood(M_c, T, X_L, X_D, T_test)

def get_column_partition_assignments(p_State):
    return p_State.get_X_L()['column_partition']['assignments']

def column_chain_to_ratio(column_chain_arr, j, i=0):
    chain_i_j = column_chain_arr[[i, j], :]
    is_same = numpy.diff(chain_i_j, axis=0)[0] == 0
    n_chains = len(is_same)
    is_same_count = sum(is_same)
    ratio = is_same_count / float(n_chains)
    return ratio

def column_partition_assignments_to_f_z_statistic(column_partition_assignments,
        j, i=0):
    iter_column_chain_arr = column_partition_assignments.transpose((1, 0, 2))
    helper = lambda column_chain_arr: column_chain_to_ratio(column_chain_arr, j, i)
    as_list = map(helper, iter_column_chain_arr)
    return numpy.array(as_list)[:, numpy.newaxis]

def default_reprocess_diagnostics_func(diagnostics_arr_dict):
    # This code formerly did stuff with the column partition
    # assignments after deleting it.  The stuff it did was apparently
    # unused, and caused trouble with one-column tables, so we have
    # removed it.  But until someone ascertains that it is safe to
    # leave the column partition assignments in, we'll continue
    # deleting it for now.
    del diagnostics_arr_dict['column_partition_assignments']
    return diagnostics_arr_dict



#
#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Lead Developers: Dan Lovell and Jay Baxter
#   Authors: Dan Lovell, Baxter Eaves, Jay Baxter, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import print_function
from six.moves import range
import copy
import itertools
import collections
import numpy
import six

import crosscat.cython_code.State as State
import crosscat.EngineTemplate as EngineTemplate
import crosscat.utils.sample_utils as su
import crosscat.utils.general_utils as gu
import crosscat.utils.inference_utils as iu

# for default_diagnostic_func_dict below
import crosscat.utils.diagnostic_utils


class LocalEngine(EngineTemplate.EngineTemplate):

    """A simple interface to the Cython-wrapped C++ engine

    LocalEngine holds no state other than a seed generator.
    Methods use resources on the local machine.

    """

    def __init__(self, seed=None):
        """Initialize a LocalEngine

        This is really just setting the initial seed to be used for
        initializing CrossCat states.  Seeds are generated sequentially

        """
        super(LocalEngine, self).__init__(seed=seed)
        self.mapper = lambda *args: list(six.moves.map(*args))
        self.do_initialize = _do_initialize_tuple
        self.do_analyze = _do_analyze_tuple
        self.do_insert = _do_insert_tuple
        return

    def get_initialize_arg_tuples(self, M_c, M_r, T, initialization,
                                  row_initialization, n_chains,
                                  ROW_CRP_ALPHA_GRID,
                                  COLUMN_CRP_ALPHA_GRID,
                                  S_GRID, MU_GRID,
                                  N_GRID,
                                  ):
        seeds = [self.get_next_seed() for seed_idx in range(n_chains)]
        arg_tuples = six.moves.zip(
            seeds,
            itertools.cycle([M_c]),
            itertools.cycle([M_r]),
            itertools.cycle([T]),
            itertools.cycle([initialization]),
            itertools.cycle([row_initialization]),
            itertools.cycle([ROW_CRP_ALPHA_GRID]),
            itertools.cycle([COLUMN_CRP_ALPHA_GRID]),
            itertools.cycle([S_GRID]),
            itertools.cycle([MU_GRID]),
            itertools.cycle([N_GRID]),
        )
        return arg_tuples

    def initialize(self, M_c, M_r, T, initialization=b'from_the_prior',
                   row_initialization=-1, n_chains=1,
                   ROW_CRP_ALPHA_GRID=(),
                   COLUMN_CRP_ALPHA_GRID=(),
                   S_GRID=(), MU_GRID=(),
                   N_GRID=31,
                   # subsample=False,
                   # subsample_proportion=None,
                   # subsample_rows_list=None,
                   ):
        """Sample a latent state from prior

        :param M_c: The column metadata
        :type M_c: dict
        :param M_r: The row metadata
        :type M_r: dict
        :param T: The data table in mapped representation (all floats, generated
                  by data_utils.read_data_objects)
        :type T: list of lists
        :returns: X_L, X_D -- the latent state

        """

        # FIXME: why is M_r passed?
        arg_tuples = self.get_initialize_arg_tuples(
            M_c, M_r, T, initialization,
            row_initialization, n_chains,
            ROW_CRP_ALPHA_GRID, COLUMN_CRP_ALPHA_GRID,
            S_GRID, MU_GRID,
            N_GRID,
        )
        chain_tuples = self.mapper(self.do_initialize, arg_tuples)

        X_L_list, X_D_list = zip(*chain_tuples)
        if n_chains == 1:
            X_L_list, X_D_list = X_L_list[0], X_D_list[0]
        return X_L_list, X_D_list

    def get_insert_arg_tuples(self, M_c, T, X_L_list, X_D_list, new_rows, N_GRID, CT_KERNEL):
        arg_tuples = six.moves.zip(
            itertools.cycle([M_c]),
            itertools.cycle([T]),
            X_L_list, X_D_list,
            itertools.cycle([new_rows]),
            itertools.cycle([N_GRID]),
            itertools.cycle([CT_KERNEL]),
        )
        return arg_tuples

    def insert(self, M_c, T, X_L_list, X_D_list, new_rows=None, N_GRID=31, CT_KERNEL=0):
        """
        Insert mutates the data T.
        """

        if new_rows is None:
            raise ValueError("new_row must exist")

        if not isinstance(new_rows, list):
            raise TypeError('new_rows must be list of lists')
            if not isinstance(new_rows[0], list):
                raise TypeError('new_rows must be list of lists')

        X_L_list, X_D_list, was_multistate = su.ensure_multistate(X_L_list, X_D_list)

        # get insert arg tuples
        arg_tuples = self.get_insert_arg_tuples(M_c, T, X_L_list, X_D_list, new_rows, N_GRID,
                                                CT_KERNEL)

        chain_tuples = self.mapper(self.do_insert, arg_tuples)
        X_L_list, X_D_list = zip(*chain_tuples)

        if not was_multistate:
            X_L_list, X_D_list = X_L_list[0], X_D_list[0]

        T.extend(new_rows)

        ret_tuple = X_L_list, X_D_list, T

        return ret_tuple

    def get_analyze_arg_tuples(self, M_c, T, X_L_list, X_D_list, kernel_list,
                               n_steps, c, r, max_iterations, max_time, diagnostic_func_dict,
                               every_N, ROW_CRP_ALPHA_GRID, COLUMN_CRP_ALPHA_GRID,
                               S_GRID, MU_GRID, N_GRID, do_timing, CT_KERNEL):
        n_chains = len(X_L_list)
        seeds = [self.get_next_seed() for seed_idx in range(n_chains)]
        arg_tuples = six.moves.zip(
            seeds,
            X_L_list, X_D_list,
            itertools.cycle([M_c]),
            itertools.cycle([T]),
            itertools.cycle([kernel_list]),
            itertools.cycle([n_steps]),
            itertools.cycle([c]),
            itertools.cycle([r]),
            itertools.cycle([max_iterations]),
            itertools.cycle([max_time]),
            itertools.cycle([diagnostic_func_dict]),
            itertools.cycle([every_N]),
            itertools.cycle([ROW_CRP_ALPHA_GRID]),
            itertools.cycle([COLUMN_CRP_ALPHA_GRID]),
            itertools.cycle([S_GRID]),
            itertools.cycle([MU_GRID]),
            itertools.cycle([N_GRID]),
            itertools.cycle([do_timing]),
            itertools.cycle([CT_KERNEL]),
        )
        return arg_tuples

    def analyze(self, M_c, T, X_L, X_D, kernel_list=(), n_steps=1, c=(), r=(),
                max_iterations=-1, max_time=-1, do_diagnostics=False,
                diagnostics_every_N=1,
                ROW_CRP_ALPHA_GRID=(),
                COLUMN_CRP_ALPHA_GRID=(),
                S_GRID=(), MU_GRID=(),
                N_GRID=31,
                do_timing=False,
                CT_KERNEL=0,
                ):
        """Evolve the latent state by running MCMC transition kernels

        :param M_c: The column metadata
        :type M_c: dict
        :param T: The data table in mapped representation (all floats, generated
                  by data_utils.read_data_objects)
        :param X_L: the latent variables associated with the latent state
        :type X_L: dict
        :param X_D: the particular cluster assignments of each row in each view
        :type X_D: list of lists
        :param kernel_list: names of the MCMC transition kernels to run
        :type kernel_list: list of strings
        :param n_steps: the number of times to run each MCMC transition kernel
        :type n_steps: int
        :param c: the (global) column indices to run MCMC transition kernels on
        :type c: list of ints
        :param r: the (global) row indices to run MCMC transition kernels on
        :type r: list of ints
        :param max_iterations: the maximum number of times ot run each MCMC
                               transition kernel. Applicable only if
                               max_time != -1.
        :type max_iterations: int
        :param max_time: the maximum amount of time (seconds) to run MCMC
                         transition kernels for before stopping to return
                         progress
        :type max_time: float
        :returns: X_L, X_D -- the evolved latent state

        """
        if n_steps <= 0:
            raise ValueError("You must do at least one analyze step.")

        if CT_KERNEL not in [0, 1]:
            raise ValueError("CT_KERNEL must be 0 (Gibbs) or 1 (MH)")

        if do_timing:
            # diagnostics and timing are exclusive
            do_diagnostics = False
        diagnostic_func_dict, reprocess_diagnostics_func = do_diagnostics_to_func_dict(
            do_diagnostics)
        X_L_list, X_D_list, was_multistate = su.ensure_multistate(X_L, X_D)
        arg_tuples = self.get_analyze_arg_tuples(M_c, T, X_L_list, X_D_list,
                                                 kernel_list, n_steps, c, r,
                                                 max_iterations, max_time,
                                                 diagnostic_func_dict, diagnostics_every_N,
                                                 ROW_CRP_ALPHA_GRID,
                                                 COLUMN_CRP_ALPHA_GRID,
                                                 S_GRID, MU_GRID,
                                                 N_GRID,
                                                 do_timing,
                                                 CT_KERNEL,
                                                 )
        chain_tuples = self.mapper(self.do_analyze, arg_tuples)
        X_L_list, X_D_list, diagnostics_dict_list = zip(*chain_tuples)
        if do_timing:
            timing_list = diagnostics_dict_list
        if not was_multistate:
            X_L_list, X_D_list = X_L_list[0], X_D_list[0]
        ret_tuple = X_L_list, X_D_list
        #
        if diagnostic_func_dict is not None:
            diagnostics_dict = munge_diagnostics(diagnostics_dict_list)
            if reprocess_diagnostics_func is not None:
                diagnostics_dict = reprocess_diagnostics_func(diagnostics_dict)
            ret_tuple = ret_tuple + (diagnostics_dict, )
        if do_timing:
            ret_tuple = ret_tuple + (timing_list, )
        return ret_tuple

    def _sample_and_insert(self, M_c, T, X_L, X_D, matching_row_indices):
        p_State = State.p_State(M_c, T, X_L, X_D)
        draws = []
        for matching_row_idx in matching_row_indices:
            random_seed = self.get_next_seed()
            draw = p_State.get_draw(matching_row_idx, random_seed)
            p_State.insert_row(draw, matching_row_idx)
            draws.append(draw)
            T.append(draw)
        X_L, X_D = p_State.get_X_L(), p_State.get_X_D()
        return draws, T, X_L, X_D

    def sample_and_insert(self, M_c, T, X_L, X_D, matching_row_idx):
        matching_row_indices = gu.ensure_listlike(matching_row_idx)
        if len(matching_row_indices) == 0:
            matching_row_indices = list(range(len(T)))
        was_single_row = len(matching_row_indices) == 1
        draws, T, X_L, X_D = self._sample_and_insert(M_c, T, X_L, X_D, matching_row_indices)
        if was_single_row:
            draws = draws[0]
        return draws, T, X_L, X_D

    def simple_predictive_sample(self, M_c, X_L, X_D, Y, Q, n=1):
        """Sample values from the predictive distribution of the given latent state

        :param M_c: The column metadata
        :type M_c: dict
        :param X_L: the latent variables associated with the latent state
        :type X_L: dict
        :param X_D: the particular cluster assignments of each row in each view
        :type X_D: list of lists
        :param Y: A list of constraints to apply when sampling.  Each constraint
                  is a triplet of (r, d, v): r is the row index, d is the column
                  index and v is the value of the constraint
        :type Y: list of lists
        :param Q: A list of values to sample.  Each value is doublet of (r, d):
                  r is the row index, d is the column index
        :type Q: list of lists
        :param n: the number of samples to draw
        :type n: int
        :returns: list of floats -- samples in the same order specified by Q

        """
        get_next_seed = self.get_next_seed
        samples = _do_simple_predictive_sample(
            M_c, X_L, X_D, Y, Q, n, get_next_seed)
        return samples

    def simple_predictive_probability(self, M_c, X_L, X_D, Y, Q):
        """Calculate the probability of a cell taking a value given a latent state

        :param M_c: The column metadata
        :type M_c: dict
        :param X_L: the latent variables associated with the latent state
        :type X_L: dict
        :param X_D: the particular cluster assignments of each row in each view
        :type X_D: list of lists
        :param Y: A list of constraints to apply when querying.  Each constraint
                  is a triplet of (r, d, v): r is the row index, d is the column
                  index and v is the value of the constraint
        :type Y: list of lists
        :param Q: A list of values to query.  Each value is triplet of (r, d, v):
                  r is the row index, d is the column index, and v is the value at
                  which the density is evaluated.
        :type Q: list of lists
        :returns: list of floats -- probabilities of the values specified by Q

        """
        return su.simple_predictive_probability(M_c, X_L, X_D, Y, Q)

    def simple_predictive_probability_multistate(self, M_c, X_L_list, X_D_list, Y, Q):
        """Calculate the probability of a cell taking a value given a latent state

        :param M_c: The column metadata
        :type M_c: dict
        :param X_L_list: list of the latent variables associated with the latent state
        :type X_L_list: list of dict
        :param X_D_list: list of the particular cluster assignments of each row in each view
        :type X_D_list: list of list of lists
        :param Y: A list of constraints to apply when querying.  Each constraint
                  is a triplet of (r,d,v): r is the row index, d is the column
                  index and v is the value of the constraint
        :type Y: list of lists
        :param Q: A list of values to query.  Each value is triplet of (r,d,v):
                  r is the row index, d is the column index, and v is the value at
                  which the density is evaluated.
        :type Q: list of lists
        :returns: list of floats -- probabilities of the values specified by Q

        """
        return su.simple_predictive_probability_multistate(M_c, X_L_list, X_D_list, Y, Q)

    def predictive_probability(self, M_c, X_L, X_D, Y, Q):
        """Calculate the probability of cellS jointly taking values given a latent state

        :param M_c: The column metadata
        :type M_c: dict
        :param X_L: the latent variables associated with the latent state
        :type X_L: dict
        :param X_D: the particular cluster assignments of each row in each view
        :type X_D: list of lists
        :param Y: A list of constraints to apply when querying.  Each constraint
                  is a triplet of (r, d, v): r is the row index, d is the column
                  index and v is the value of the constraint
        :type Y: list of lists
        :param Q: A list of values to query.  Each value is triplet of (r, d, v):
                  r is the row index, d is the column index, and v is the value at
                  which the density is evaluated.
        :type Q: list of lists
        :returns: float -- joint log probability of the values specified by Q

        """
        return su.predictive_probability(M_c, X_L, X_D, Y, Q)

    def predictive_probability_multistate(self, M_c, X_L_list, X_D_list, Y, Q):
        """Calculate the probability of cellS jointly taking values given a latent state

        :param M_c: The column metadata
        :type M_c: dict
        :param X_L_list: list of the latent variables associated with the latent state
        :type X_L_list: list of dict
        :param X_D_list: list of the particular cluster assignments of each row in each view
        :type X_D_list: list of list of lists
        :param Y: A list of constraints to apply when querying.  Each constraint
                  is a triplet of (r,d,v): r is the row index, d is the column
                  index and v is the value of the constraint
        :type Y: list of lists
        :param Q: A list of values to query.  Each value is triplet of (r,d,v):
                  r is the row index, d is the column index, and v is the value at
                  which the density is evaluated.
        :type Q: list of lists
        :returns: float -- joint log probabilities of the values specified by Q

        """
        return su.predictive_probability_multistate(M_c, X_L_list, X_D_list, Y, Q)

    def mutual_information(self, M_c, X_L_list, X_D_list, Q, n_samples=1000):
        """
        Return the estimated mutual information for each pair of columns on Q given
        the set of samples.

        :param M_c: The column metadata
        :type M_c: dict
        :param X_L_list: list of the latent variables associated with the latent state
        :type X_L_list: list of dict
        :param X_D_list: list of the particular cluster assignments of each row in each view
        :type X_D_list: list of list of lists
        :param Q: List of tuples where each tuple contains the two column indexes to compare
        :type Q: list of two-tuples of ints
        :param n_samples: the number of simple predictive samples to use
        :type n_samples: int
        :returns: list of list, where each sublist is a set of MIs and Linfoots from each crosscat
        sample.
        """
        return iu.mutual_information(M_c, X_L_list, X_D_list, Q, n_samples)

    def row_structural_typicality(self, X_L_list, X_D_list, row_id):
        """
        Returns the typicality (opposite of anomalousness) of the given row.

        :param X_L_list: list of the latent variables associated with the latent state
        :type X_L_list: list of dict
        :param X_D_list: list of the particular cluster assignments of each row in each view
        :type X_D_list: list of list of lists
        :param row_id: id of the target row
        :type row_id: int
        :returns: float, the typicality, from 0 to 1
        """
        return su.row_structural_typicality(X_L_list, X_D_list, row_id)

    def column_structural_typicality(self, X_L_list, col_id):
        """
        Returns the typicality (opposite of anomalousness) of the given column.

        :param X_L_list: list of the latent variables associated with the latent state
        :type X_L_list: list of dict
        :param col_id: id of the target col
        :type col_id: int
        :returns: float, the typicality, from 0 to 1
        """
        return su.column_structural_typicality(X_L_list, col_id)

    def similarity(self, M_c, X_L_list, X_D_list, given_row_id, target_row_id, target_columns=None):
        """Computes the similarity of the given row to the target row, averaged over all the
        column indexes given by target_columns.

        :param M_c: The column metadata
        :type M_c: dict
        :param X_L: list of the latent variables associated with the latent state
        :type X_L: list of dicts
        :param X_D: list of the particular cluster assignments of each row in each view
        :type X_D: list of list of lists
        :param given_row_id: the id of one of the rows to measure similarity between
        :type given_row_id: int
        :param target_row_id: the id of the other row to measure similarity between
        :type target_row_id: int
        :param target_columns: the columns to average the similarity over. defaults to all columns.
        :type target_columns: int, string, or list of ints
        :returns: float

        """
        return su.similarity(M_c, X_L_list, X_D_list, given_row_id, target_row_id, target_columns)

    def impute(self, M_c, X_L, X_D, Y, Q, n):
        """Impute values from the predictive distribution of the given latent state

        :param M_c: The column metadata
        :type M_c: dict
        :param X_L: the latent variables associated with the latent state
        :type X_L: dict
        :param X_D: the particular cluster assignments of each row in each view
        :type X_D: list of lists
        :param Y: A list of constraints to apply when sampling.  Each constraint
                  is a triplet of (r,d,v): r is the row index, d is the column
                  index and v is the value of the constraint
        :type Y: list of lists
        :param Q: A list of values to sample.  Each value is doublet of (r, d):
                  r is the row index, d is the column index
        :type Q: list of lists
        :param n: the number of samples to use in the imputation
        :type n: int
        :returns: list of floats -- imputed values in the same order as
                  specified by Q

        """
        e = su.impute(M_c, X_L, X_D, Y, Q, n, self.get_next_seed)
        return e

    def impute_and_confidence(self, M_c, X_L, X_D, Y, Q, n):
        """Impute values and confidence of the value from the predictive
        distribution of the given latent state

        :param M_c: The column metadata
        :type M_c: dict
        :param X_L: the latent variables associated with the latent state
        :type X_L: dict
        :param X_D: the particular cluster assignments of each row in each view
        :type X_D: list of lists
        :param Y: A list of constraints to apply when sampling.  Each constraint
                  is a triplet of (r, d, v): r is the row index, d is the column
                  index and v is the value of the constraint
        :type Y: list of lists
        :param Q: A list of values to sample.  Each value is doublet of (r, d):
                  r is the row index, d is the column index
        :type Q: list of lists
        :param n: the number of samples to use in the imputation
        :type n: int
        :returns: list of lists -- list of (value, confidence) tuples in the
                  same order as specified by Q

        """
        if isinstance(X_L, (list, tuple)):
            assert isinstance(X_D, (list, tuple))
            # TODO: multistate impute doesn't exist yet
            # e,confidence = su.impute_and_confidence_multistate(M_c, X_L, X_D, Y, Q, n,
            #                                                    self.get_next_seed)
            e, confidence = su.impute_and_confidence(
                M_c, X_L, X_D, Y, Q, n, self.get_next_seed)
        else:
            e, confidence = su.impute_and_confidence(
                M_c, X_L, X_D, Y, Q, n, self.get_next_seed)
        return (e, confidence)

    def ensure_col_dep_constraints(self, M_c, M_r, T, X_L, X_D, dep_constraints,
            max_rejections=100):
        """Ensures dependencey or indepdendency between columns.

        dep_constraints is a list of where each entry is an (int, int, bool) tuple
        where the first two entries are column indices and the third entry
        describes whether the columns are to be dependent (True) or independent
        (False).

        Behavior Notes:
        ensure_col_dep_constraints will add col_esnure enforcement to the
        metadata (top level of X_L); unensure_col will remove it. Calling
        ensure_col_dep_constraints twice will replace the first ensure.

        This operation destroys the existing X_L and X_D metadata; the user
        should be aware that it will clobber any existing analyses.

        Implementation Notes:
        Initialization is implemented via rejection (by repeatedly initalizing
        states and throwing ones out that do not adhear to dep_constraints).
        This means that in the event the contraints in dep_constraints are
        complex, or impossible, that the rejection alogrithm may fail.

        The returned metadata looks like this:
        >>> dep_constraints
        [(1, 2, True), (2, 5, True), (1, 5, True), (1, 3, False)]
        >>> X_L['col_ensure']
        {
            "dependent" :
            {
                1 : [2, 5],
                2 : [1, 5],
                5 : [1, 2]
            },
            "independent" :
            {
                1 : [3],
                3 : [1]
        }
        """
        X_L_list, X_D_list, was_multistate = su.ensure_multistate(X_L, X_D)
        if was_multistate:
            num_states = len(X_L_list)
        else:
            num_states = 1

        col_ensure_md = dict()
        col_ensure_md[True] = dict()
        col_ensure_md[False] = dict()

        for col1, col2, dependent in dep_constraints:
            if col1 == col2:
                raise ValueError("Cannot specify same columns in dependence"\
                    " constraints.")
            if str(col1) in col_ensure_md[dependent]:
                col_ensure_md[dependent][str(col1)].append(col2)
            else:
                col_ensure_md[dependent][str(col1)] = [col2]
            if col2 in col_ensure_md[dependent]:
                col_ensure_md[dependent][str(col2)].append(col1)
            else:
                col_ensure_md[dependent][str(col2)] = [col1]

        def assert_dep_constraints(X_L, X_D, dep_constraints):
            for col1, col2, dep in dep_constraints:
                if not self.assert_col_dep_constraints(X_L, X_D, col1, col2,
                    dep, True):
                    return False
            return True

        X_L_out = []
        X_D_out = []
        for _ in range(num_states):
            counter = 0
            X_L_i, X_D_i = self.initialize(M_c, M_r, T)
            while not assert_dep_constraints(X_L_i, X_D_i, dep_constraints):
                if counter > max_rejections:
                    raise RuntimeError("Could not ranomly generate a partition"\
                        " that satisfies the constraints in dep_constraints.")
                counter += 1
                X_L_i, X_D_i = self.initialize(M_c, M_r, T)

            X_L_i['col_ensure'] = dict()
            X_L_i['col_ensure']['dependent'] = col_ensure_md[True]
            X_L_i['col_ensure']['independent'] = col_ensure_md[False]

            X_D_out.append(X_D_i)
            X_L_out.append(X_L_i)

        if was_multistate:
            return X_L_out, X_D_out
        else:
            return X_L_out[0], X_D_out[0]

    def ensure_row_dep_constraint(self, M_c, T, X_L, X_D, row1, row2,
            dependent=True, wrt=None, max_iter=100, force=False):
        """Ensures dependencey or indepdendency between rows with respect to
        (wrt) columns."""
        X_L_list, X_D_list, was_multistate = su.ensure_multistate(X_L, X_D)
        if force:
            raise NotImplementedError
        else:
            kernel_list = ('row_partition_assignements',)
            for i, (X_L_i, X_D_i) in enumerate(zip(X_L_list, X_D_list)):
                iters = 0
                X_L_tmp = copy.deepcopy(X_L_i)
                X_D_tmp = copy.deepcopy(X_D_i)
                while not self.assert_row(X_L_tmp, X_D_tmp, row1, row2,
                        dependent=dependent, wrt=wrt):
                    if iters >= max_iter:
                        raise RuntimeError('Maximum ensure iterations reached.')
                    res = self.analyze(M_c, T, X_L_i, X_D_i, kernel_list=kernel_list,
                        n_steps=1, r=(row1,))
                    X_L_tmp = res[0]
                    X_D_tmp = res[1]
                    iters += 1
                X_L_list[i] = X_L_tmp
                X_D_list[i] = X_D_tmp

        if was_multistate:
            return X_L_list, X_D_list
        else:
            return X_L_list[0], X_D_list[0]

    def assert_col_dep_constraints(self, X_L, X_D, col1, col2, dependent=True,
        single_bool=False):
        # TODO: X_D is not used for anything other than ensure_multistate.
        # I should probably edit ensure_multistate to take X_L or X_D using
        # keyword arguments.
        X_L_list, _, was_multistate = su.ensure_multistate(X_L, X_D)
        model_assertions = []
        assertion = True
        for X_L_i in X_L_list:
            assg = X_L_i['column_partition']['assignments']
            assertion = (assg[col1] == assg[col2]) == dependent
            if single_bool and not assertion:
                return False
            model_assertions.append(assertion)

        if single_bool:
            return True

        if was_multistate:
            return model_assertions
        else:
            return model_assertions[0]

    def assert_row(self, X_L, X_D, row1, row2, dependent=True, wrt=None):
        X_L_list, X_D_list, was_multistate = su.ensure_multistate(X_L, X_D)
        if wrt is None:
            num_cols = len(X_L_list[0]['column_partition']['assignments'])
            wrt = list(range(num_cols))
        else:
            if not isinstance(wrt, list):
                raise TypeError('wrt must be a list')
        model_assertions = []
        for X_L_i, X_D_i in zip(X_L_list, X_D_list):
            view_assg = X_L_i['column_partition']['assignments']
            views_wrt = list(set([view_assg[col] for col in wrt]))
            model_assertion = True
            for view in views_wrt:
                if (X_D_i[view][row1] == X_D_i[view][row2]) != dependent:
                    model_assertion = False
                    break
            model_assertions.append(model_assertion)

        if was_multistate:
            return model_assertions
        else:
            return model_assertions[0]
        pass


def do_diagnostics_to_func_dict(do_diagnostics):
    diagnostic_func_dict = None
    reprocess_diagnostics_func = None
    if do_diagnostics:
        if isinstance(do_diagnostics, (dict,)):
            diagnostic_func_dict = do_diagnostics
        else:
            diagnostic_func_dict = dict(default_diagnostic_func_dict)
        if 'reprocess_diagnostics_func' in diagnostic_func_dict:
            reprocess_diagnostics_func = diagnostic_func_dict.pop(
                'reprocess_diagnostics_func')
    return diagnostic_func_dict, reprocess_diagnostics_func


def get_value_in_each_dict(key, dict_list):
    return numpy.array([dict_i[key] for dict_i in dict_list]).T


def munge_diagnostics(diagnostics_dict_list):
    # all dicts should have the same keys
    diagnostic_names = diagnostics_dict_list[0].keys()
    diagnostics_dict = {
        diagnostic_name: get_value_in_each_dict(diagnostic_name, diagnostics_dict_list)
        for diagnostic_name in diagnostic_names
    }
    return diagnostics_dict

# switched ordering so args that change come first
# FIXME: change LocalEngine.initialze to match ordering here


def _do_initialize(SEED, M_c, M_r, T, initialization, row_initialization,
                   ROW_CRP_ALPHA_GRID, COLUMN_CRP_ALPHA_GRID,
                   S_GRID, MU_GRID,
                   N_GRID,
                   ):
    p_State = State.p_State(M_c, T, initialization=initialization,
                            row_initialization=row_initialization, SEED=SEED,
                            ROW_CRP_ALPHA_GRID=ROW_CRP_ALPHA_GRID,
                            COLUMN_CRP_ALPHA_GRID=COLUMN_CRP_ALPHA_GRID,
                            S_GRID=S_GRID,
                            MU_GRID=MU_GRID,
                            N_GRID=N_GRID,
                            )
    X_L = p_State.get_X_L()
    X_D = p_State.get_X_D()
    return X_L, X_D


def _do_initialize_tuple(arg_tuple):
    return _do_initialize(*arg_tuple)


def _do_insert_tuple(arg_tuple):
    return _do_insert(*arg_tuple)


def _do_insert(M_c, T, X_L, X_D, new_rows, N_GRID, CT_KERNEL):
    p_State = State.p_State(M_c, T, X_L=X_L, X_D=X_D,
                            N_GRID=N_GRID,
                            CT_KERNEL=CT_KERNEL)

    row_idx = len(T)
    for row_data in new_rows:
        p_State.insert_row(row_data, row_idx)
        p_State.transition(which_transitions=['row_partition_assignments'], r=[row_idx])
        row_idx += 1

    X_L_prime = p_State.get_X_L()
    X_D_prime = p_State.get_X_D()
    return X_L_prime, X_D_prime

# switched ordering so args that change come first
# FIXME: change LocalEngine.analyze to match ordering here


def _do_analyze(SEED, X_L, X_D, M_c, T, kernel_list, n_steps, c, r,
                max_iterations, max_time,
                ROW_CRP_ALPHA_GRID, COLUMN_CRP_ALPHA_GRID,
                S_GRID, MU_GRID,
                N_GRID,
                CT_KERNEL,
                ):
    p_State = State.p_State(M_c, T, X_L, X_D, SEED=SEED,
                            ROW_CRP_ALPHA_GRID=ROW_CRP_ALPHA_GRID,
                            COLUMN_CRP_ALPHA_GRID=COLUMN_CRP_ALPHA_GRID,
                            S_GRID=S_GRID,
                            MU_GRID=MU_GRID,
                            N_GRID=N_GRID,
                            CT_KERNEL=CT_KERNEL
                            )
    p_State.transition(kernel_list, n_steps, c, r,
                       max_iterations, max_time)
    X_L_prime = p_State.get_X_L()
    X_D_prime = p_State.get_X_D()
    return X_L_prime, X_D_prime


def _do_analyze_tuple(arg_tuple):
    return _do_analyze_with_diagnostic(*arg_tuple)


def get_child_n_steps_list(n_steps, every_N):
    if every_N is None:
        # results in one block of size n_steps
        every_N = n_steps
    missing_endpoint = numpy.arange(0, n_steps, every_N)
    with_endpoint = numpy.append(missing_endpoint, n_steps)
    child_n_steps_list = numpy.diff(with_endpoint)
    return child_n_steps_list.tolist()

none_summary = lambda p_State: None

# switched ordering so args that change come first
# FIXME: change LocalEngine.analyze to match ordering here


def _do_analyze_with_diagnostic(SEED, X_L, X_D, M_c, T, kernel_list, n_steps, c, r,
                                max_iterations, max_time, diagnostic_func_dict, every_N,
                                ROW_CRP_ALPHA_GRID, COLUMN_CRP_ALPHA_GRID,
                                S_GRID, MU_GRID,
                                N_GRID,
                                do_timing,
                                CT_KERNEL,
                                ):
    diagnostics_dict = collections.defaultdict(list)
    if diagnostic_func_dict is None:
        diagnostic_func_dict = dict()
        every_N = None
    child_n_steps_list = get_child_n_steps_list(n_steps, every_N)
    # import ipdb; ipdb.set_trace()
    p_State = State.p_State(M_c, T, X_L, X_D, SEED=SEED,
                            ROW_CRP_ALPHA_GRID=ROW_CRP_ALPHA_GRID,
                            COLUMN_CRP_ALPHA_GRID=COLUMN_CRP_ALPHA_GRID,
                            S_GRID=S_GRID,
                            MU_GRID=MU_GRID,
                            N_GRID=N_GRID,
                            CT_KERNEL=CT_KERNEL,
                            )
    with gu.Timer('all transitions', verbose=False) as timer:
        for child_n_steps in child_n_steps_list:
            p_State.transition(kernel_list, child_n_steps, c, r,
                               max_iterations, max_time)
            for diagnostic_name, diagnostic_func in six.iteritems(diagnostic_func_dict):
                diagnostic_value = diagnostic_func(p_State)
                diagnostics_dict[diagnostic_name].append(diagnostic_value)
                pass
            pass
        pass
    X_L_prime = p_State.get_X_L()
    X_D_prime = p_State.get_X_D()
    #
    if do_timing:
        # diagnostics and timing are exclusive
        diagnostics_dict = timer.elapsed_secs
        pass
    return X_L_prime, X_D_prime, diagnostics_dict


def _do_simple_predictive_sample(M_c, X_L, X_D, Y, Q, n, get_next_seed):
    is_multistate = su.get_is_multistate(X_L, X_D)
    if is_multistate:
        samples = su.simple_predictive_sample_multistate(M_c, X_L, X_D, Y, Q,
                                                         get_next_seed, n)
    else:
        samples = su.simple_predictive_sample(M_c, X_L, X_D, Y, Q,
                                              get_next_seed, n)
    return samples


default_diagnostic_func_dict = dict(
    # fully qualify path b/c dview.sync_imports can't deal with 'as'
    # imports
    logscore=crosscat.utils.diagnostic_utils.get_logscore,
    num_views=crosscat.utils.diagnostic_utils.get_num_views,
    column_crp_alpha=crosscat.utils.diagnostic_utils.get_column_crp_alpha,
    # any outputs required by reproess_diagnostics_func must be generated
    # as well
    column_partition_assignments=crosscat.utils.diagnostic_utils.get_column_partition_assignments,
    reprocess_diagnostics_func=crosscat.utils.diagnostic_utils.default_reprocess_diagnostics_func,
)


if __name__ == '__main__':
    import crosscat.tests.timing_test_utils as ttu
    import crosscat.utils.data_utils as du
    import crosscat.utils.convergence_test_utils as ctu

    # settings
    gen_seed = 0
    inf_seed = 0
    num_clusters = 4
    num_cols = 32
    num_rows = 400
    num_views = 2
    n_steps = 1
    n_times = 5
    n_chains = 3
    n_test = 100
    CT_KERNEL = 1

    # generate some data
    T, M_r, M_c, data_inverse_permutation_indices = du.gen_factorial_data_objects(
        gen_seed, num_clusters, num_cols, num_rows, num_views,
        max_mean=100, max_std=1, send_data_inverse_permutation_indices=True)
    view_assignment_truth, X_D_truth = ctu.truth_from_permute_indices(
        data_inverse_permutation_indices, num_rows, num_cols, num_views, num_clusters)

    # run some tests
    engine = LocalEngine(seed=inf_seed)
    multi_state_ARIs = []
    multi_state_mean_test_lls = []
    X_L_list, X_D_list = engine.initialize(M_c, M_r, T, n_chains=n_chains)
    multi_state_ARIs.append(
        ctu.get_column_ARIs(X_L_list, view_assignment_truth))

    for time_i in range(n_times):
        X_L_list, X_D_list = engine.analyze(
            M_c, T, X_L_list, X_D_list, n_steps=n_steps, CT_KERNEL=CT_KERNEL)
        multi_state_ARIs.append(
            ctu.get_column_ARIs(X_L_list, view_assignment_truth))
        # multi_state_mean_test_lls.append(
        #     ctu.calc_mean_test_log_likelihoods(M_c, T,
        #                                        X_L_list, X_D_list, T_test))

    X_L_list, X_D_list, diagnostics_dict = engine.analyze(
        M_c, T, X_L_list, X_D_list,
        n_steps=n_steps, do_diagnostics=True)

    # print results
    ct_kernel_name = 'UNKNOWN'
    if CT_KERNEL == 0:
        ct_kernel_name = 'GIBBS'
    elif CT_KERNEL == 1:
        ct_kernel_name = 'METROPOLIS'

    print('Running with %s CT_KERNEL' % ct_kernel_name)
    print('generative_mean_test_log_likelihood')
    # print generative_mean_test_log_likelihood
    #
    print('multi_state_mean_test_lls:')
    print(multi_state_mean_test_lls)
    #
    print('multi_state_ARIs:')
    print(multi_state_ARIs)



import operator
import itertools
from functools import partial
#
import crosscat.tests.geweke_utils as geweke_utils
from crosscat.utils.general_utils import MapperContext, NoDaemonPool, Timer


def _generate_args_list(num_rows, num_iters, ct_kernel, cctypes):
    num_cols = len(cctypes)
    args_list = [
            '--num_rows', str(num_rows),
            '--num_cols', str(num_cols),
            '--CT_KERNEL', str(ct_kernel),
            '--num_iters', str(num_iters),
            '--cctypes'
            ] + cctypes
    return args_list

def _gen_cctypes(*args):
    _cctypes = [[cctype] * N for (cctype, N) in args]
    return reduce(operator.add, _cctypes)

def generate_args_list(base_num_rows, num_iters, do_long=False, _divisor=1.):
    num_cols_list = [1, 10]
    col_type_list = ['continuous', 'multinomial']
    col_type_pairs = sorted(itertools.combinations(col_type_list, 2))
    ct_kernel_list = [0, 1]
    args_list = []

    # single datatype
    iter_over = itertools.product(col_type_list, num_cols_list, ct_kernel_list)
    for col_type, num_cols, ct_kernel in iter_over:
        cctypes = _gen_cctypes((col_type, num_cols))
        args = _generate_args_list(base_num_rows, num_iters, ct_kernel, cctypes)
        args += ['--_divisor', str(_divisor)]
        args_list.append(args)
        pass

    # pairs of datatypes
    iter_over = itertools.product(col_type_pairs, num_cols_list, ct_kernel_list)
    for (col_type_a, col_type_b), num_cols, ct_kernel in iter_over:
        cctypes = _gen_cctypes((col_type_a, num_cols), (col_type_b, num_cols))
        args = _generate_args_list(base_num_rows, num_iters, ct_kernel, cctypes)
        args += ['--_divisor', str(_divisor)]
        args_list.append(args)
        pass

    # hard coded runs
    if do_long:
        num_cols_long = 100
        num_rows_long = 100
        #
        cctypes = _gen_cctypes(('continuous', num_cols_long))
        args = _generate_args_list(num_rows_long, num_iters, cctypes)
        args += ['--_divisor', str(_divisor)]
        args_list.append(args + ['--CT_KERNEL', str(0)])
        args_list.append(args + ['--CT_KERNEL', str(1)])
        #
        cctypes = _gen_cctypes(('multinomial', num_cols_long))
        args = _generate_args_list(num_rows_long, num_iters, cctypes)
        args += ['--_divisor', str(_divisor)]
        args += ['--num_multinomial_values', '2']
        args_list.append(args + ['--CT_KERNEL', str(0)])
        args_list.append(args + ['--CT_KERNEL', str(1)])
        #
#        cctypes = _gen_cctypes(('multinomial', num_cols_long))
#        args = _generate_args_list(num_rows_long, num_iters, cctypes)
#        args += ['--num_multinomial_values', '128']
#        args_list.append(args)
        pass
    return args_list

def plot_results(results, dirname='./'):
    with Timer('plotting') as timer:
        with MapperContext(Pool=NoDaemonPool) as mapper:
            # use non-daemonic mapper since plot_result spawns daemonic processes
            plotter = partial(geweke_utils.plot_result,
                    dirname=dirname)
            mapper(plotter, results)
            pass
        pass
    return

if __name__ == '__main__':
    import argparse
    import experiment_runner.experiment_utils as eu
    from experiment_runner.ExperimentRunner import ExperimentRunner, propagate_to_s3
    parser = argparse.ArgumentParser()
    parser.add_argument('--dirname', default='geweke_on_schemas', type=str)
    parser.add_argument('--base_num_rows', default=10, type=int)
    parser.add_argument('--num_iters', default=2000, type=int)
    parser.add_argument('--no_plots', action='store_true')
    parser.add_argument('--do_long', action='store_true')
    parser.add_argument('--_divisor', default=1., type=float)

    args = parser.parse_args()
    dirname = args.dirname
    base_num_rows = args.base_num_rows
    num_iters = args.num_iters
    do_plots = not args.no_plots
    do_long = args.do_long
    _divisor = args._divisor


    # create configs
    arg_list_to_config = partial(eu.arg_list_to_config,
            geweke_utils.generate_parser(),
            arbitrate_args=geweke_utils.arbitrate_args)
    args_list = generate_args_list(base_num_rows, num_iters, do_long, _divisor)
    config_list = map(arg_list_to_config, args_list)


    # do experiments
    er = ExperimentRunner(geweke_utils.run_geweke, storage_type='fs',
            dirname_prefix=dirname,
            bucket_str='experiment_runner')
    er.do_experiments(config_list)
    # push to s3
    propagate_to_s3(er)


    if do_plots:
        for id in er.frame.index:
            result = er._get_result(id)
            geweke_utils.plot_result(result, dirname)
            pass
        pass

    print(er.frame)



#
#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Lead Developers: Dan Lovell and Jay Baxter
#   Authors: Dan Lovell, Baxter Eaves, Jay Baxter, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import print_function
import numpy
import crosscat.cython_code.ContinuousComponentModel as CCM
import crosscat.cython_code.MultinomialComponentModel as MCM
import crosscat.cython_code.State as State

c_hypers = dict(r=10,nu=10,s=10,mu=10)
ccm = CCM.p_ContinuousComponentModel(c_hypers)
print("empty component model")
print(ccm)
#
for element in [numpy.nan, 0, 1, numpy.nan, 2]:
    print()
    ccm.insert_element(element)
    print("inserted %s" % element)
    print(ccm)

m_hypers = dict(dirichlet_alpha=10,K=3)
mcm = MCM.p_MultinomialComponentModel(m_hypers)
print("empty component model")
print(mcm)

for element in [numpy.nan, 0, 1, numpy.nan, 2]:
    print()
    mcm.insert_element(element)
    print("inserted %s" % element)
    print(mcm)



#
#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Lead Developers: Dan Lovell and Jay Baxter
#   Authors: Dan Lovell, Baxter Eaves, Jay Baxter, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import print_function
from libcpp cimport bool
from libcpp.vector cimport vector
from libcpp.string cimport string
from libcpp.map cimport map as c_map
from libcpp.set cimport set as c_set
from cython.operator import dereference
cimport numpy as np
#
import numpy
import six
#
# import crosscat.tests.plot_utils as pu
import crosscat.utils.file_utils as fu
import crosscat.utils.general_utils as gu


cdef c_map[int, c_set[int]] empty_map_of_int_set():
    cdef c_map[int, c_set[int]] retval
    return retval

cdef double set_double(double& to_set, double value):
     (&to_set)[0] = value	# (&...)[0] works around Cython >=0.18 bug.
     return to_set

cdef vector[int] convert_int_vector_to_cpp(python_vector):
     cdef vector[int] ret_vec
     for value in python_vector:
          ret_vec.push_back(value)
     return ret_vec

cdef vector[string] convert_string_vector_to_cpp(python_vector):
     cdef vector[string] ret_vec
     cdef string s
     for value in python_vector:
          s = value if isinstance(value, six.binary_type) else value.encode()
          ret_vec.push_back(s)
     return ret_vec

cdef extern from "<boost/numeric/ublas/matrix.hpp>" namespace "boost::numeric::ublas":
    cdef cppclass matrix[double]:
        void clear()
        int size1()
        int size2()
        double& operator()(int i, int j)
    matrix[double] *new_matrix "new boost::numeric::ublas::matrix<double>" (int i, int j)
    void del_matrix "delete" (matrix *m)

cdef matrix[double]* convert_data_to_cpp(np.ndarray[np.float64_t, ndim=2] data):
     cdef int num_rows = data.shape[0]
     cdef int num_cols = data.shape[1]
     dataptr = new_matrix(num_rows, num_cols)
     cdef int i,j
     for i from 0 <= i < num_rows:
          for j from 0 <= j < num_cols:
               set_double(dereference(dataptr)(i,j), data[i,j])
     return dataptr

cdef extern from "State.h":
     cdef cppclass State:
          # mutators
          double insert_row(vector[double] row_data, int matching_row_idx, int
                  row_idx)
          double transition(matrix[double] data)
          double transition_column_crp_alpha()
          double transition_features(matrix[double] data, vector[int] which_cols)
          double transition_column_hyperparameters(vector[int] which_cols)
          double transition_row_partition_hyperparameters(vector[int] which_cols)
          double transition_row_partition_assignments(matrix[double] data, vector[int] which_rows)
          double transition_views(matrix[double] data)
          double transition_view_i(int i, matrix[double] data)
          double transition_views_row_partition_hyper()
          double transition_views_col_hypers()
          double transition_views_zs(matrix[double] data)
          double calc_row_predictive_logp(vector[double] in_vd)
          # getters
          double get_column_crp_alpha()
          double get_column_crp_score()
          double get_data_score()
          double get_marginal_logp()
          vector[double] get_draw(int row_idx, int random_seed)
          int get_num_views()
          c_map[int, vector[int]] get_column_groups()
          string to_string(string join_str, bool top_level)
          double draw_rand_u()
          int draw_rand_i()
          # API helpers
          vector[c_map[string, double]] get_column_hypers()
          c_map[string, double] get_column_partition_hypers()
          vector[int] get_column_partition_assignments()
          vector[int] get_column_partition_counts()
          #
          c_map[string, double] get_row_partition_model_hypers_i(int view_idx)
          c_map[int, c_set[int]] get_column_dependencies()
          c_map[int, c_set[int]] get_column_independencies()
          vector[int] get_row_partition_model_counts_i(int view_idx)
          vector[vector[c_map[string, double]]] get_column_component_suffstats_i(int view_idx)
          #
          vector[vector[int]] get_X_D()
          void SaveResult()
     State *new_State "new State" (matrix[double] &data,
                                   vector[string] global_col_datatypes,
                                   vector[int] global_col_multinomial_counts,
                                   vector[int] global_row_indices,
                                   vector[int] global_col_indices,
                                   string col_initialization,
                                   string row_initialization,
                                   vector[double] ROW_CRP_ALPHA_GRID,
                                   vector[double] COLUMN_CRP_ALPHA_GRID,
                                   vector[double] S_GRID,
                                   vector[double] MU_GRID,
                                   int N_GRID, int SEED, int CT_KERNEL)
     State *new_State "new State" (matrix[double] &data,
                                   vector[string] global_col_datatypes,
                                   vector[int] global_col_multinomial_counts,
                                   vector[int] global_row_indices,
                                   vector[int] global_col_indices,
                                   c_map[int, c_map[string, double]] hypers_m,
                                   vector[vector[int]] column_partition,
                                   c_map[int, c_set[int]] col_ensure_dep,
                                   c_map[int, c_set[int]] col_ensure_ind,
                                   double column_crp_alpha,
                                   vector[vector[vector[int]]] row_partition_v,
                                   vector[double] row_crp_alpha_v,
                                   vector[double] ROW_CRP_ALPHA_GRID,
                                   vector[double] COLUMN_CRP_ALPHA_GRID,
                                   vector[double] S_GRID,
                                   vector[double] MU_GRID,
                                   int N_GRID, int SEED, int CT_KERNEL)
     void del_State "delete" (State *s)


def extract_column_types_counts(M_c):
    column_types = [
        column_metadata['modeltype']
        for column_metadata in M_c['column_metadata']
        ]
    event_counts = [
        len(column_metadata.get('value_to_code',[]))
        for column_metadata in M_c['column_metadata']
        ]
    return column_types, event_counts

def get_args_dict(args_list, vars_dict):
     args_dict = dict([
               (arg, vars_dict[arg])
               for arg in args_list
               ])
     return args_dict

transition_name_to_method_name_and_args = dict(
     column_partition_hyperparameter=('transition_column_crp_alpha', []),
     column_partition_assignments=('transition_features', ['c']),
     column_hyperparameters=('transition_column_hyperparameters', ['c']),
     row_partition_hyperparameters= ('transition_row_partition_hyperparameters', ['c']),
     row_partition_assignments=('transition_row_partition_assignments', ['r']),
     )

def get_all_transitions_permuted(seed):
     which_transitions = transition_name_to_method_name_and_args.keys()
     random_state = numpy.random.RandomState(seed)
     which_transitions = random_state.permutation(which_transitions)
     return which_transitions

cdef class p_State:
    cdef State *thisptr
    cdef matrix[double] *dataptr
    cdef vector[int] gri
    cdef vector[int] gci
    cdef vector[string] column_types
    cdef vector[int] event_counts
    cdef np.ndarray T_array
    cpdef M_c

    def __cinit__(self, M_c, T, X_L=None, X_D=None,
                  initialization='from_the_prior', row_initialization=-1,
                  ROW_CRP_ALPHA_GRID=(), COLUMN_CRP_ALPHA_GRID=(),
                  S_GRID=(), MU_GRID=(),
                  N_GRID=31, SEED=0, CT_KERNEL=0):
         column_types, event_counts = extract_column_types_counts(M_c)
         global_row_indices = range(len(T))
         global_col_indices = range(len(T[0]))
         #
         # FIXME: keeping TWO copies of the data here
         self.T_array = numpy.array(T)
         self.dataptr = convert_data_to_cpp(self.T_array)
         self.column_types = convert_string_vector_to_cpp(column_types)
         self.event_counts = convert_int_vector_to_cpp(event_counts)
         self.gri = convert_int_vector_to_cpp(global_row_indices)
         self.gci = convert_int_vector_to_cpp(global_col_indices)
         self.M_c = M_c
         #
         must_initialize = X_L is None
         if must_initialize:
              col_initialization = initialization
              if row_initialization == -1:
                   row_initialization = initialization
              self.thisptr = new_State(dereference(self.dataptr),
                                       self.column_types,
                                       self.event_counts,
                                       self.gri, self.gci,
                                       col_initialization,
                                       row_initialization,
                                       ROW_CRP_ALPHA_GRID,
                                       COLUMN_CRP_ALPHA_GRID,
                                       S_GRID, MU_GRID,
                                       N_GRID, SEED, CT_KERNEL)
         else:
              # # !!! MUTATES X_L !!!
              desparsify_X_L(M_c, X_L)
              constructor_args = \
                  transform_latent_state_to_constructor_args(X_L, X_D)
              hypers_m = constructor_args['hypers_m']
              column_partition = constructor_args['column_partition']
              column_crp_alpha = constructor_args['column_crp_alpha']
              row_partition_v = constructor_args['row_partition_v']
              row_crp_alpha_v = constructor_args['row_crp_alpha_v']
              col_ensure_dep = constructor_args['col_ensure_dep']
              col_ensure_ind = constructor_args['col_ensure_ind']
              if col_ensure_dep is None:
                  col_ensure_dep = empty_map_of_int_set()
                  col_ensure_ind = empty_map_of_int_set()
              self.thisptr = new_State(dereference(self.dataptr),
                                       self.column_types,
                                       self.event_counts,
                                       self.gri, self.gci,
                                       hypers_m,
                                       column_partition,
                                       col_ensure_dep,
                                       col_ensure_ind,
                                       column_crp_alpha,
                                       row_partition_v, row_crp_alpha_v,
                                       ROW_CRP_ALPHA_GRID,
                                       COLUMN_CRP_ALPHA_GRID,
                                       S_GRID, MU_GRID,
                                       N_GRID, SEED, CT_KERNEL)
    def __dealloc__(self):
        del_matrix(self.dataptr)
        del_State(self.thisptr)
    def __repr__(self):
         print_tuple = (
              self.dataptr.size1(),
              self.dataptr.size2(),
              self.thisptr.to_string(";", False),
              )
         return "State[%s, %s]:\n%s" % print_tuple
    def to_string(self, join_str='\n', top_level=False):
         return self.thisptr.to_string(join_str, top_level)
    # def plot(self, M_c, filename=None, dir=''):
    #      T_array = self.T_array
    #      X_D = self.get_X_D()
    #      X_L = self.get_X_L()
    #      pu.plot_views(T_array, X_D, X_L, M_c, filename, dir)
    # def plot_T(self, M_c, filename=None, dir=''):
    #      T_array = self.T_array
    #      pu.plot_T(T_array, M_c, filename, dir)
    #
    # getters
    def get_column_groups(self):
        return self.thisptr.get_column_groups()
    def get_column_crp_alpha(self):
         return self.thisptr.get_column_crp_alpha()
    def get_column_crp_score(self):
         return self.thisptr.get_column_crp_score()
    def get_data_score(self):
         return self.thisptr.get_data_score()
    def get_marginal_logp(self):
        return self.thisptr.get_marginal_logp()
    def get_num_views(self):
        return self.thisptr.get_num_views()
    def calc_row_predictive_logp(self, in_vd):
         return self.thisptr.calc_row_predictive_logp(in_vd)
    def get_draw(self, row_idx, random_seed):
        return self.thisptr.get_draw(row_idx, random_seed)
    #
    # get_X_L helpers helpers
    def get_row_partition_model_i(self, view_idx):
          hypers = self.thisptr.get_row_partition_model_hypers_i(view_idx)
          counts = self.thisptr.get_row_partition_model_counts_i(view_idx)
          row_partition_model_i = dict()
          row_partition_model_i['hypers'] = hypers
          row_partition_model_i['counts'] = counts
          return row_partition_model_i
    def get_column_names_i(self, view_idx):
         idx_to_name = self.M_c['idx_to_name']
         column_groups = self.thisptr.get_column_groups()
         #
         column_indices_i = column_groups[view_idx]
         column_indices_i = map(str, column_indices_i)
         column_names_i = []
         for idx in column_indices_i:
              if idx not in idx_to_name:
                   print("%r not in %r" % (idx, idx_to_name))
              value = idx_to_name[idx]
              column_names_i.append(value)
         return column_names_i
    def get_column_component_suffstats_i(self, view_idx):
         column_component_suffstats = \
             self.thisptr.get_column_component_suffstats_i(view_idx)
         # FIXME: make this actually work
         #        should sparsify here rather than later in get_X_L
         # column_names = self.get_column_names_i(view_idx)
         # for col_name, column_component_suffstats_i in \
         #          zip(column_names, column_component_suffstats):
         #     modeltype = get_modeltype_from_name(self.M_c, col_name)
         #     if modeltype == 'symmetric_dirichlet_discrete':
         #         sparsify_column_component_suffstats(column_component_suffstats_i)
         return column_component_suffstats
    def get_view_state_i(self, view_idx):
          row_partition_model = self.get_row_partition_model_i(view_idx)
          column_names = self.get_column_names_i(view_idx)
          column_component_suffstats = \
              self.get_column_component_suffstats_i(view_idx)
          view_state_i = dict()
          view_state_i['row_partition_model'] = row_partition_model
          view_state_i['column_names'] = list(column_names)
          view_state_i['column_component_suffstats'] = \
              column_component_suffstats
          return view_state_i
    # get_X_L helpers
    def get_column_partition(self):
        hypers = self.thisptr.get_column_partition_hypers()
        assignments = self.thisptr.get_column_partition_assignments()
        counts = self.thisptr.get_column_partition_counts()
        column_partition = dict()
        column_partition['hypers'] = hypers
        column_partition['assignments'] = assignments
        column_partition['counts'] = counts
        return column_partition
    def get_column_hypers(self):
        return self.thisptr.get_column_hypers()
    def get_view_state(self):
        view_state = []
        for view_idx in range(self.get_num_views()):
            view_state_i = self.get_view_state_i(view_idx)
            view_state.append(view_state_i)
        return view_state

    def get_col_ensure_dep(self):
        retval = self.thisptr.get_column_dependencies()
        if len(retval) == 0:
            return None
        else:
            return retval

    def get_col_ensure_ind(self):
        retval = self.thisptr.get_column_independencies()
        if len(retval) == 0:
            return None
        else:
            return retval

    # mutators
    def insert_row(self, row_data, matching_row_idx, row_idx=-1):
        return self.thisptr.insert_row(row_data, matching_row_idx, row_idx)
    def transition(self, which_transitions=(), n_steps=1,
                   c=(), r=(), max_iterations=-1, max_time=-1):
         seed = None
         score_delta = 0
         if len(which_transitions) == 0:
              seed = self.thisptr.draw_rand_i()
              which_transitions = get_all_transitions_permuted(seed)
         with gu.Timer('transition', verbose=False) as timer:
              for step_idx in range(n_steps):
                   for which_transition in which_transitions:
                        elapsed_secs = timer.get_elapsed_secs()
                        if (max_time != -1) and (max_time < elapsed_secs):
                             break
                        method_name_and_args = transition_name_to_method_name_and_args.get(which_transition)
                        if method_name_and_args is not None:
                             method_name, args_list = method_name_and_args
                             which_method = getattr(self, method_name)
                             args_dict = get_args_dict(args_list, locals())
                             score_delta += which_method(**args_dict)
                        else:
                             print_str = 'INVALID TRANSITION TYPE TO ' \
                                 'State.transition: %s' % which_transition
                             print(print_str)
         return score_delta
    def transition_column_crp_alpha(self):
         return self.thisptr.transition_column_crp_alpha()
    def transition_features(self, c=()):
        return self.thisptr.transition_features(dereference(self.dataptr), c)
    def transition_column_hyperparameters(self, c=()):
         return self.thisptr.transition_column_hyperparameters(c)
    def transition_row_partition_hyperparameters(self, c=()):
         return self.thisptr.transition_row_partition_hyperparameters(c)
    def transition_row_partition_assignments(self, r=()):
         return self.thisptr.transition_row_partition_assignments(dereference(self.dataptr), r)
    def transition_views(self):
        return self.thisptr.transition_views(dereference(self.dataptr))
    def transition_view_i(self, i):
        return self.thisptr.transition_view_i(i, dereference(self.dataptr))
    def transition_views_col_hypers(self):
         return self.thisptr.transition_views_col_hypers()
    def transition_views_row_partition_hyper(self):
         return self.thisptr.transition_views_row_partition_hyper()
    def transition_views_zs(self):
         return self.thisptr.transition_views_zs(dereference(self.dataptr))
    # API getters
    def get_X_D(self):
          return self.thisptr.get_X_D()
    def get_X_L(self):
          column_partition = self.get_column_partition()
          column_hypers = self.get_column_hypers()
          view_state = self.get_view_state()

          # Need to convert from c_map[int c_set[int] to dict(string:list).
          col_ensure_dep = self.get_col_ensure_dep()
          col_ensure_ind = self.get_col_ensure_ind()

          if col_ensure_dep is None:
             col_ensure_dep_json = {}
          else:
             col_ensure_dep_json = {str(k):list(v) for (k,v) in col_ensure_dep.items()}
          if col_ensure_ind is None:
             col_ensure_ind_json = {}
          else:
             col_ensure_ind_json = {str(k):list(v) for (k,v) in col_ensure_ind.items()}

          X_L = dict()
          X_L['column_partition'] = column_partition
          X_L['column_hypers'] = column_hypers
          X_L['view_state'] = view_state
          if col_ensure_dep is not None or col_ensure_ind is not None:
              X_L['col_ensure'] = dict()
              X_L['col_ensure']['dependent'] = col_ensure_dep_json
              X_L['col_ensure']['independent'] = col_ensure_ind_json

          sparsify_X_L(self.M_c, X_L)
          return X_L
    def save(self, filename, dir='', **kwargs):
         save_dict = dict(
              X_L=self.get_X_L(),
              X_D=self.get_X_D(),
              )
         save_dict.update(**kwargs)
         fu.pickle(save_dict, filename, dir=dir)

def indicator_list_to_list_of_list(indicator_list):
     list_of_list = []
     num_clusters = max(indicator_list) + 1
     import numpy
     for cluster_idx in range(num_clusters):
          which_rows = numpy.array(indicator_list)==cluster_idx
          list_of_list.append(numpy.nonzero(which_rows)[0])
     return list_of_list

def floatify_dict(in_dict):
     for key in in_dict:
          in_dict[key] = float(in_dict[key])
     return in_dict

def floatify_dict_dict(in_dict):
     for key in in_dict:
          in_dict[key] = floatify_dict(in_dict[key])
     return in_dict

def extract_row_partition_alpha(view_state):
     hypers = view_state['row_partition_model']['hypers']
     alpha = hypers.get('alpha')
     if alpha is None:
          log_alpha = hypers['log_alpha']
          alpha = numpy.exp(log_alpha)
     return alpha

def transform_latent_state_to_constructor_args(X_L, X_D):
     num_rows = len(X_D[0])
     num_cols = len(X_L['column_hypers'])
     #
     global_row_indices = range(num_rows)
     global_col_indices = range(num_cols)
     hypers_m = dict(zip(global_col_indices, X_L['column_hypers']))
     hypers_m = floatify_dict_dict(hypers_m)
     column_indicator_list = X_L['column_partition']['assignments']
     column_partition = indicator_list_to_list_of_list(column_indicator_list)
     column_crp_alpha = X_L['column_partition']['hypers']['alpha']
     row_partition_v = map(indicator_list_to_list_of_list, X_D)
     row_crp_alpha_v = map(extract_row_partition_alpha, X_L['view_state'])

     # Need to convert from dict(string:list) to c_map[int c_set[int].
     if X_L.get('col_ensure', None) is None:
         col_ensure_dep = empty_map_of_int_set()
         col_ensure_ind = empty_map_of_int_set()
     else:
         col_ensure_dep_json = X_L['col_ensure'].get('dependent', None)
         if col_ensure_dep_json is None:
           col_ensure_dep = empty_map_of_int_set()
         else:
           col_ensure_dep = {int(k):set(v) for (k,v) in col_ensure_dep_json.items()}
         col_ensure_ind_json = X_L['col_ensure'].get('independent', None)
         if col_ensure_ind_json is None:
           col_ensure_ind = empty_map_of_int_set()
         else:
           col_ensure_ind = {int(k):set(v) for (k,v) in col_ensure_ind_json.items()}

     n_grid = 31
     seed = 0
     ct_kernel=0
     #
     constructor_args = dict()
     constructor_args['global_row_indices'] = global_row_indices
     constructor_args['global_col_indices'] = global_col_indices
     constructor_args['hypers_m'] = hypers_m
     constructor_args['column_partition'] = column_partition
     constructor_args['column_crp_alpha'] = column_crp_alpha
     constructor_args['row_partition_v'] = row_partition_v
     constructor_args['row_crp_alpha_v'] = row_crp_alpha_v
     constructor_args['N_GRID'] = n_grid
     constructor_args['SEED'] = seed
     constructor_args['CT_KERNEL'] = ct_kernel
     constructor_args['col_ensure_dep'] = col_ensure_dep
     constructor_args['col_ensure_ind'] = col_ensure_ind
     #
     return constructor_args

def without_zero_values(dict_in):
    return {k: v for k, v in six.iteritems(dict_in) if v != 0}

def insert_zero_values(dict_in, N_keys):
     for key in range(N_keys):
          key_str = str(key)
          if key_str not in dict_in:
               dict_in[key_str] = 0.0
     return dict_in

def sparsify_column_component_suffstats(column_component_suffstats):
     for idx, suffstats_i in enumerate(column_component_suffstats):
          suffstats_i = without_zero_values(suffstats_i)
          column_component_suffstats[idx] = suffstats_i
     return None

def desparsify_column_component_suffstats(column_component_suffstats,
                                          N_keys):
     for idx, suffstats_i in enumerate(column_component_suffstats):
          insert_zero_values(suffstats_i, N_keys)
     return None

def get_column_component_suffstats_by_global_col_idx(M_c, X_L, col_idx):
     col_name = M_c['idx_to_name'][str(col_idx)]
     view_idx = X_L['column_partition']['assignments'][col_idx]
     view_state_i = X_L['view_state'][view_idx]
     within_view_idx = view_state_i['column_names'].index(col_name)
     column_component_suffstats_i = view_state_i['column_component_suffstats'][within_view_idx]
     return column_component_suffstats_i
     
def sparsify_X_L(M_c, X_L):
     for col_idx, col_i_metadata in enumerate(M_c['column_metadata']):
          modeltype = col_i_metadata['modeltype']
          if modeltype != 'symmetric_dirichlet_discrete':
               continue
          column_component_suffstats_i = \
              get_column_component_suffstats_by_global_col_idx(M_c, X_L,
                                                               col_idx)
          sparsify_column_component_suffstats(column_component_suffstats_i)
     return None

def desparsify_X_L(M_c, X_L):
     for col_idx, col_i_metadata in enumerate(M_c['column_metadata']):
          modeltype = col_i_metadata['modeltype']
          if modeltype != 'symmetric_dirichlet_discrete':
               continue
          column_component_suffstats_i = \
              get_column_component_suffstats_by_global_col_idx(M_c, X_L,
                                                               col_idx)
          N_keys = len(col_i_metadata['value_to_code'])
          desparsify_column_component_suffstats(column_component_suffstats_i,
                                                N_keys)
     return None

def get_modeltype_from_name(M_c, col_name):
    global_col_idx = M_c['name_to_idx'][col_name]
    modeltype = M_c['column_metadata'][global_col_idx]['modeltype']
    return modeltype



#
#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Lead Developers: Dan Lovell and Jay Baxter
#   Authors: Dan Lovell, Baxter Eaves, Jay Baxter, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from libcpp.vector cimport vector
from libcpp.string cimport string as cpp_string
from libcpp.map cimport map as cpp_map
from cython.operator import dereference


cdef extern from "string" namespace "std":
	cdef cppclass string:
		char* c_str()
		string(char*)

cdef cpp_string get_string(in_string):
	cdef cpp_string cps = string(in_string)
	return cps

cdef set_string_double_map(cpp_map[cpp_string, double] &out_map, in_map):
	for key in in_map:
		out_map[get_string(key)] = in_map[key]

cdef extern from "CyclicComponentModel.h":
	cdef cppclass CyclicComponentModel:
		cpp_map[cpp_string, double] _get_suffstats()
		cpp_map[cpp_string, double] get_hypers()
		int get_count()
		double score
		cpp_string to_string()
		double get_draw(int seed)
		double get_draw_constrained(int seed, vector[double] constraints)
		double get_predictive_pdf(double element, vector[double] constraints)
		double insert_element(double element)
		double remove_element(double element)
		double incorporate_hyper_update()
		double calc_marginal_logp()
		double calc_element_predictive_logp(double element)
		double calc_element_predictive_logp_constrained(double element, vector[double] constraints)
	CyclicComponentModel *new_CyclicComponentModel "new CyclicComponentModel" (cpp_map[cpp_string, double] &in_hypers)
	CyclicComponentModel *new_CyclicComponentModel "new CyclicComponentModel" (cpp_map[cpp_string, double] &in_hypers, int COUNT, double SUM_SIN_X, double SUM_COS_X)
	void del_CyclicComponentModel "delete" (CyclicComponentModel *ccm)

cdef class p_CyclicComponentModel:
	cdef CyclicComponentModel *thisptr
	cdef cpp_map[cpp_string, double] hypers
	def __cinit__(self, in_map, count=None, sum_sin_x=None, sum_cos_x=None):
		set_string_double_map(self.hypers, in_map)
		if count is None:
			self.thisptr = new_CyclicComponentModel(self.hypers)
		else:
			self.thisptr = new_CyclicComponentModel(self.hypers,
				count, sum_sin_x, sum_cos_x)
	def __dealloc__(self):
		del_CyclicComponentModel(self.thisptr)
	def get_draw(self, seed):
		return self.thisptr.get_draw(seed)
	def get_draw_constrained(self, seed, constraints):
		return self.thisptr.get_draw_constrained(seed, constraints)
	def get_hypers(self):
		return self.thisptr.get_hypers()
	def get_suffstats(self):
		return self.thisptr._get_suffstats()
	def get_count(self):
		return self.thisptr.get_count()
	def insert_element(self, element):
		return self.thisptr.insert_element(element)
	def remove_element(self, element):
		return self.thisptr.remove_element(element)
	def incorporate_hyper_update(self):
		return self.thisptr.incorporate_hyper_update()
	def calc_marginal_logp(self):
		return self.thisptr.calc_marginal_logp()
	def calc_element_predictive_logp(self, element):
		return self.thisptr.calc_element_predictive_logp(element)
	def calc_element_predictive_logp_constrained(self, element, constraints):
		return self.thisptr.calc_element_predictive_logp_constrained(element, constraints)
	def __repr__(self):
		return self.thisptr.to_string()



#
#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Lead Developers: Dan Lovell and Jay Baxter
#   Authors: Dan Lovell, Baxter Eaves, Jay Baxter, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import print_function
from six.moves import cPickle
from six.moves import range
import os
import re
import argparse
import zlib
import base64
#
import crosscat.settings as S
from crosscat.settings import Hadoop as hs
import crosscat.utils.file_utils as fu
import crosscat.utils.data_utils as du


default_table_data_filename = hs.default_table_data_filename
default_table_filename = hs.default_table_filename
default_analyze_args_dict = hs.default_analyze_args_dict.copy()
default_initialize_args_dict = hs.default_initialize_args_dict.copy()


# read the data, create metadata
def read_and_pickle_table_data(table_data_filename, pkl_filename):
    T, M_r, M_c = du.read_model_data_from_csv(table_data_filename,
                                              gen_seed=0)
    table_data = dict(T=T, M_r=M_r, M_c=M_c)
    fu.pickle(table_data, pkl_filename)
    return table_data

# cat the data into the script, as hadoop would
def run_script_local(infile, script_name, outfile, table_data_filename=None):
    infix_str = ''
    if table_data_filename is not None:
        infix_str = '--table_data_filename %s' % table_data_filename
    cmd_str = 'cat %s | python %s %s > %s'
    cmd_str %= (infile, script_name, infix_str, outfile)
    print(cmd_str)
    os.system(cmd_str)
    return

def my_dumps(in_object):
    ret_str = cPickle.dumps(in_object)
    ret_str = zlib.compress(ret_str)
    ret_str = base64.b64encode(ret_str) 
    return ret_str

def my_loads(in_str):
    in_str = base64.b64decode(in_str)
    in_str = zlib.decompress(in_str)
    out_object = cPickle.loads(in_str)
    return out_object

def write_hadoop_line(fh, key, dict_to_write):
    escaped_str = my_dumps(dict_to_write)
    fh.write(str(key) + ' ')
    fh.write(escaped_str)
    fh.write('\n')
    return

line_re = '(\d+)\s+(.*)'
pattern = re.compile(line_re)
def parse_hadoop_line(line):
    line = line.strip()
    match = pattern.match(line)
    key, dict_in = None, None
    if match:
        key, dict_in_str = match.groups()
        try:
          dict_in = my_loads(dict_in_str)
        except Exception as e:
          # for parsing new NLineInputFormat
          match = pattern.match(dict_in_str)
          if match is None:
            print('OMG: ' + dict_in_str[:50])
            import pdb; pdb.set_trace()
          key, dict_in_str = match.groups()
          dict_in = my_loads(dict_in_str)
    return key, dict_in

def write_support_files(table_data, table_data_filename,
                        command_dict, command_dict_filename):
    fu.pickle(table_data, table_data_filename)
    fu.pickle(command_dict, command_dict_filename)
    return

def write_initialization_files(initialize_input_filename,
                               table_data, table_data_filename,
                               initialize_args_dict, intialize_args_dict_filename,
                               n_chains=10):
    write_support_files(table_data, table_data_filename,
                        initialize_args_dict, intialize_args_dict_filename)
    with open(initialize_input_filename, 'w') as out_fh:
        for SEED in range(n_chains):
            out_dict = dict(SEED=SEED)
            write_hadoop_line(out_fh, SEED, out_dict)
    return

def write_analyze_files(analyze_input_filename, X_L_list, X_D_list,
                        table_data, table_data_filename,
                        analyze_args_dict, analyze_args_dict_filename,
                        SEEDS=None):
    assert len(X_L_list) == len(X_D_list)
    write_support_files(table_data, table_data_filename,
                        analyze_args_dict, analyze_args_dict_filename)
    if SEEDS is None:
        SEEDS = range(len(X_L_list))
    with open(analyze_input_filename, 'w') as out_fh:
        for SEED, X_L, X_D in zip(SEEDS, X_L_list, X_D_list):
            out_dict = dict(SEED=SEED, X_L=X_L, X_D=X_D)
            write_hadoop_line(out_fh, SEED, out_dict)
    return

# read initialization output, write analyze input
def link_initialize_to_analyze(initialize_output_filename,
                               analyze_input_filename,
                               analyze_args_dict=None):
    if analyze_args_dict is None:
        analyze_args_dict = default_analyze_args_dict.copy()
    num_lines = 0
    with open(initialize_output_filename) as in_fh:
        with open(analyze_input_filename, 'w') as out_fh:
            for line in in_fh:
                num_lines += 1
                key, dict_in = parse_hadoop_line(line)
                dict_in.update(analyze_args_dict)
                dict_in['SEED'] = int(key)
                write_hadoop_line(out_fh, key, dict_in)
    return num_lines

def get_is_vpn_connected():
    # cmd_str = 'ifconfig | grep tun'
    cmd_str = 'ping -W 2 -c 1 10.1.90.10'
    lines = [line for line in os.popen(cmd_str)]
    is_vpn_connected = False
    if len(lines) != 0:
        is_vpn_connected = True
    return is_vpn_connected

def assert_vpn_is_connected():
    is_vpn_connected = get_is_vpn_connected()
    assert is_vpn_connected
    return


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('do_what', type=str)
    parser.add_argument('--hadoop_filename', type=str, default=None)
    parser.add_argument('--table_filename',
        default=default_table_filename, type=str)
    parser.add_argument('--pkl_filename',
                        default=default_table_data_filename, type=str)
    parser.add_argument('--initialize_input_filename',
                        default='initialize_input', type=str)
    parser.add_argument('--initialize_output_filename',
                        default='initialize_output', type=str)
    parser.add_argument('--analyze_input_filename',
                        default='analyze_input', type=str)
    parser.add_argument('--n_steps', default=1, type=int)
    parser.add_argument('--n_chains', default=1, type=int)
    args = parser.parse_args()
    #
    do_what = args.do_what
    hadoop_filename = args.hadoop_filename
    table_filename = args.table_filename
    pkl_filename = args.pkl_filename
    initialize_input_filename = args.initialize_input_filename
    initialize_output_filename = args.initialize_output_filename
    analyze_input_filename = args.analyze_input_filename
    n_steps = args.n_steps
    n_chains = args.n_chains


    if do_what == 'read_and_pickle_table_data':
        read_and_pickle_table_data(table_filename, pkl_filename)
    elif do_what == 'write_initialization_files':
        write_initialization_files(initialize_input_filename,
                                   n_chains=n_chains)
    elif do_what == 'link_initialize_to_analyze':
        analyze_args_dict = default_analyze_args_dict.copy()
        analyze_args_dict['n_steps'] = n_steps
        link_initialize_to_analyze(initialize_output_filename,
                                   analyze_input_filename,
                                   analyze_args_dict)
    elif do_what == 'assert_vpn_is_connected':
        assert_vpn_is_connected()
    elif do_what == 'parse_hadoop_lines':
        assert hadoop_filename is not None
        parsed_lines = []
        with open(hadoop_filename) as fh:
            for line in fh:
                parsed_lines.append(parse_hadoop_line(line))
                print(len(parsed_lines))
        if pkl_filename != default_table_data_filename:
            fu.pickle(parsed_lines, pkl_filename)
    else:
        print('unknown do_what: %s' % do_what)



#
#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Lead Developers: Dan Lovell and Jay Baxter
#   Authors: Dan Lovell, Baxter Eaves, Jay Baxter, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import argparse
#
import numpy
#
import crosscat.utils.data_utils as du
import crosscat.cython_code.State as State
from crosscat.JSONRPCEngine import JSONRPCEngine


parser = argparse.ArgumentParser()
parser.add_argument('--gen_seed', default=0, type=int)
parser.add_argument('--inf_seed', default=0, type=int)
parser.add_argument('--num_clusters', default=4, type=int)
parser.add_argument('--num_cols', default=16, type=int)
parser.add_argument('--num_rows', default=300, type=int)
parser.add_argument('--num_splits', default=2, type=int)
parser.add_argument('--max_mean', default=10, type=float)
parser.add_argument('--max_std', default=0.3, type=float)
parser.add_argument('--num_transitions', default=30, type=int)
parser.add_argument('--N_GRID', default=31, type=int)
parser.add_argument('--URI', default='http://localhost:8007', type=str)
args = parser.parse_args()
#
gen_seed = args.gen_seed
inf_seed = args.inf_seed
num_clusters = args.num_clusters
num_cols = args.num_cols
num_rows = args.num_rows
num_splits = args.num_splits
max_mean = args.max_mean
max_std = args.max_std
num_transitions = args.num_transitions
N_GRID = args.N_GRID
URI = args.URI


# create the data
T, M_r, M_c = du.gen_factorial_data_objects(
    gen_seed, num_clusters,
    num_cols, num_rows, num_splits,
    max_mean=max_mean, max_std=max_std,
    )

#
engine = JSONRPCEngine(inf_seed, URI=URI)

# initialize
X_L, X_D = engine.initialize(M_c, M_r, T)

# analyze without do_diagnostics or do_timing
X_L, X_D = engine.analyze(M_c, T, X_L, X_D, n_steps=num_transitions)

# analyze with do_diagnostics
X_L, X_D, diagnostics_dict = engine.analyze(M_c, T, X_L, X_D, n_steps=num_transitions, do_diagnostics=True)

# analyze with do_timing
X_L, X_D, timing_list = engine.analyze(M_c, T, X_L, X_D, n_steps=num_transitions, do_timing=True)

## draw sample states
#for sample_idx in range(num_samples):
#    print "starting sample_idx #: %s" % sample_idx
#    X_L, X_D = engine.analyze(M_c, T, X_L, X_D, kernel_list, lag,
#                              c, r, max_iterations, max_time)
#    p_State = State.p_State(M_c, T, X_L, X_D, N_GRID=N_GRID)
#    plot_filename = 'sample_%s_X_D' % sample_idx
#    pkl_filename = 'sample_%s_pickled_state.pkl.gz' % sample_idx
#    p_State.save(filename=pkl_filename, M_c=M_c, T=T)
#    p_State.plot(filename=plot_filename)



#
#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Lead Developers: Dan Lovell and Jay Baxter
#   Authors: Dan Lovell, Baxter Eaves, Jay Baxter, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import print_function
#
import ContinuousComponentModel as CCM


hyper_map = dict()
hyper_map["mu"] = 1.0
hyper_map["s"] = 1.0
hyper_map["nu"] = 1.0
hyper_map["r"] = 2.0

CCM.set_string_double_map(hyper_map)
component_model = CCM.p_ContinuousComponentModel(hyper_map)
print(component_model.calc_marginal_logp())
component_model.insert_element(2.3)
print(component_model.calc_marginal_logp())
print(component_model)
print("component_model.get_draw(0):", component_model.get_draw(0))
print("component_model.get_draw(1):", component_model.get_draw(1))
component_model.remove_element(2.3)
print(component_model.calc_marginal_logp())



#
#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Lead Developers: Dan Lovell and Jay Baxter
#   Authors: Dan Lovell, Baxter Eaves, Jay Baxter, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# A set of utilities that help with generating random states and calculating
# their probabilities.
# Functions list: 
#  1.GenerateRandomState(): Generates a random state and fills with gaussian data
#  2.GenerateRandomPartition(): Generates a random state (partitions only)
#  3.CRP(): Generates an assignment vector according to the CRP
#  4.pflip(): flips an n-dimensional hypercoin
#  5.GenDataFromPartitions(): generates data adhearing to a partitioning scheme
#  6.Bell(): Returns the Bell number
#  7.Stirling2nd(): Returns the Stirling number of the 2nd kind, s(n,k)
#  8.lcrp(): returns the log probability of a partition under the CRP
#  9.NGML(): calculates the Normal-Gamma marginal likelihood
# 10.CCML(): Calculates the marginal likelihood of a continuous data array under CrossCat
# 11.FixPriors(): Changes the priors in X_L
# 12.vectorToCHMat(): converts Assignment vector to cohabitation matrix (for state matching)
# 13.(class) CrossCatPartitions: Partition object. Enumerates all states.
# 14.(class) Partition: Used to enumerate partitions for CrossCatPartitions

from __future__ import print_function
from time import time

import math

import random as rand
import numpy as np
import scipy as sp
import itertools

from scipy.special import binom

import crosscat.cython_code.State as State
import crosscat.utils.data_utils as du


# generates from a state with the columns to views partition col_parts, and 
# the rows in views to cats partition row_parts (X_D). Accepts partitions
# as numpy array or lists. Returns a state; the data table, T; and M_c
# Arguments:
#	col_parts : vector of length n_cols assigning columns to views
#	row_parts : a list of vectors (or numpy array) with n_views rows where
#				each row, i, is a vector assigning the rows of the columns 
#				in view i to categories
#	mean_gen  : the mean of the means of data clusters
#	std_gen   : the standard deviation of cluster means
#	std_data  : the standard deviation of the individual clusters
def GenerateStateFromPartitions(col_parts, row_parts, mean_gen=0.0, std_gen=1.0, std_data=0.1):
	
	T, M_r, M_c = GenDataFromPartitions(col_parts, row_parts, mean_gen=mean_gen, std_gen=std_gen, std_data=std_data)
	state = State.p_State(M_c, T, N_GRID=100)

	X_L = state.get_X_L()
	X_D = state.get_X_D()

	if type(col_parts) is not list:
		X_L['column_partition']['assignments'] = col_parts.tolist()

	if type(row_parts) is not list:
		X_D = row_parts.tolist()

	# create a new state with the updated X_D and X_L
	state = State.p_State(M_c, T, X_L=X_L, X_D=X_D, N_GRID=100)

	return state, T, M_c, M_r, X_L, X_D

# generates a random state with n_rows rows and n_cols columns, fills it with 
# normal data and the specified alphas, and prepares it for running. Returns 
# a State.p_state object and the table of data and M_c
# Arguments:
#	n_rows    : the number of rows
#	n_cols    : the number of columns
#	mean_gen  : the mean of the means of data clusters
#	std_gen   : the standard deviation of cluster means
#	std_data  : the standard deviation of the individual clusters
#	alpha_col : the CRP parameter for columns to views
# 	alpha_rows: the CRP parameter for rows into categories
def GenerateRandomState(n_rows, n_cols, mean_gen=0.0, std_gen=1.0, std_data=0.1, alpha_col=1.0, alpha_rows=1.0):

	# check the inputs 
	assert(type(n_rows) is int)
	assert(type(n_cols) is int)
	assert(type(mean_gen) is float)
	assert(type(std_gen) is float)
	assert(type(std_data) is float)
	assert(type(alpha_col) is float)
	assert(type(alpha_rows) is float)
	assert(n_rows > 0)
	assert(n_cols > 0)
	assert(std_gen > 0.0)
	assert(std_data > 0.0)
	assert(alpha_col > 0.0)
	assert(alpha_rows > 0.0)

	# generate the partitioning
	part = GenerateRandomPartition(n_rows, n_cols, alpha_col, alpha_rows)

	# fill it with data
	T, M_r, M_c = GenDataFromPartitions(part['col_parts'], part['row_parts'], mean_gen, std_gen, std_data)

	# this part is kind of hacky:
	# generate a state from the prior 
	state = State.p_State(M_c, T, N_GRID=100)
	# get the X_L and X_D and implant part['col_parts'], part['row_parts'], then 
	# create a new state with the new X_L and X_D defined
	X_L = state.get_X_L()
	X_D = state.get_X_D()

	# this should be all we need to change for 
	# State.transform_latent_state_to_constructor_args(X_L, X_D) to be able
	# to construct the arguments to intialize a state
	X_L['column_partition']['assignments'] = part['col_parts'].tolist()
	X_D = part['row_parts'].tolist()

	# hack in the alpha values supplied (or not) by the user
	X_L['column_partition']['hypers']['alpha'] = alpha_col
	for i in range(len(X_L['view_state'])):
		X_L['view_state'][i]['row_partition_model']['hypers']['alpha'] = alpha_col
	for i in range(n_cols):
		X_L['column_hypers'][i]['alpha'] = alpha_rows

	# create a new state with the updated X_D and X_L
	state = State.p_State(M_c, T, X_L=X_L, X_D=X_D, N_GRID=100)

	return state, T, M_r, M_c

# generates a random partitioning of n_rows rows and n_cols columns based on the
# CRP. The resulting partition is a dict with two entries: ['col_parts'] and
# ['row_parts']. See CrossCatPartitions for details on these.
def GenerateRandomPartition(n_rows, n_cols, alpha_col=1.0, alpha_rows=1.0):
	assert(type(n_rows) is int)
	assert(type(n_cols) is int)
	assert(type(alpha_col) is float)
	assert(type(alpha_rows) is float)
	assert(n_rows > 0)
	assert(n_cols > 0)
	assert(alpha_col > 0.0)
	assert(alpha_rows > 0.0)

	column_partition = CRP(n_cols, alpha_col)
	n_views = max(column_partition)+1;	

	row_partition = np.zeros(shape=(n_views,n_rows),dtype=int)

	for i in range(n_views):
		row_partition_tmp = CRP(n_rows, alpha_rows)
		row_partition[i] = row_partition_tmp

	partition = dict();
	partition['col_parts'] = column_partition;
	partition['row_parts'] = row_partition;

	assert(len(column_partition)==n_cols)
	assert(row_partition.shape[0]==n_views)
	assert(row_partition.shape[1]==n_rows)

	return partition

# Generates an N-length partitioning through the Chinese Restauraunt Process 
# with discount parameter alpha
def CRP(N,alpha):
	assert(type(N) is int)
	assert(type(alpha) is float)
	assert(N > 0)
	assert(alpha > 0.0)

	partition = np.zeros(N,dtype=int);

	for i in range(1,N):
		K = max(partition[0:i])+1;
		ps = np.zeros(K+1)
		for k in range(K):
			# get the number of people sitting at table k
			Nk = np.count_nonzero(partition[0:i]==k);
			ps[k] = Nk/(i+alpha)
		
		ps[K] = alpha/(i+alpha)

		assignment = pflip(ps)

		partition[i] = assignment

	assert(len(partition)==N)

	return partition

# flips an n-sided hypercoin
def pflip(P):
	# seed the RNG with system time
	rand.seed(None)
	if type(P) is float:
		if P > 1 or P < 0:
			print("Error: pflip: P is a single value not in [0,1]. P=" + str(P))
		else:
			return 1 if rand.random() > .5 else 0
	elif type(P) is np.ndarray:
		# normalize the list
		P = P/sum(P)
		P = np.cumsum(P)
		rdex = rand.random()
		# return the first entry greater than rdex	
		return np.nonzero(P>rdex)[0][0]
	else:
		print("Error: pflip: P is an invalid type.")

# Generates T, M_c, and M_r fitting the state defined by col_part and row_part.
# Generates only continuous data.
def GenDataFromPartitions(col_part,row_parts,mean_gen,std_gen,std_data):
	n_cols = len(col_part)
	n_rows = row_parts.shape[1]

	seed = int(time()*100)
	np.random.seed(seed)

	T = np.zeros((n_rows,n_cols))

	for col in range(n_cols):
		view = col_part[col]
		row_part = row_parts[view,:]
		cats = max(row_part)+1
		for cat in range(cats):
			row_dex = np.nonzero(row_part==cat)[0]
			n_rows_cat = len(row_dex)
			mean = np.random.normal(mean_gen,std_gen)
			X = np.random.normal(mean,std_data,(n_rows_cat,1))
			i = 0
			for row in row_dex:
				T[row,col] = X[i]
				i += 1

	
	T = T.tolist()
	M_r = du.gen_M_r_from_T(T)
	M_c = du.gen_M_c_from_T(T)

	return T, M_r, M_c

# CrossCatPartitions 
# enumerates all states with n_rows rows and n_cols columns. This should not be 
# used for anything bigger than 4-by-4. If you want to generate large states, 
# generate them randomly using GenerateRandomPartition
#
# Attributes:
#	.N 		- The number of states
# 	.states - A list of state dicts. Each state has:
#		['idx'] 	 : an integer index of the state
#		['col_parts']: a n_cols length vector assigning columns to views.
#			Ex. state['col_parts'] = [0 0 1] implies that columns 0 and 1 are 
#				assigned to view 0 and column 2 is assigned to view 1.
#		['row_parts']: a n_views-length list of n_rows-length vectors. 
#			Each entry, i, assigns the row to a category. 
#			Ex. state['row_parts'] = [ [0 0 1], [0 0 0] ] implies that there are
#				two views. In the first view, the rows 0 and 1 are assigned to 
#				category 0 and row 2 is assigned to category 1. In the second 
#				view, all rows are assigned to category 0.
class CrossCatPartitions(object):

	def __init__(self,n_rows, n_cols):
		# Holds the total number of partitionings
		self.N = 0;
		Bn = Bell(n_rows);

		# Generate the column partitons (views)
		self.col_partition = Partition.EnumeratePartitions(n_cols)

		# Generate the possible row partition
		self.row_partition = Partition.EnumeratePartitions(n_rows)

		# each entry of the list row_perms holds the permutation of 
		# row_partition for for each view. So if row_perm[1][2] = (0,2) it means
		# that there are two views in the second view  partitioning and that in 
		# the third permutation that the rows in view 0 are partitions according
		# to self.row_partition[0] and that the rows in the second view are 
		# partitioned according to self.row_partition[2]
		self.row_perms = [];

		# for each partition, generate the partitioning of the rows of each view 
		# into categories (cells)
		for i in range(0,self.col_partition.shape[1]):
			# get the number of partitions
			K = i+1
			r = list(range(1, int(Bn)+1))

			# generate the permutations with replacement
			perms = [];
			for t in itertools.product(r, repeat = K):
				perms.append(t)

			self.row_perms.append(perms)


		for i in range(0,self.col_partition.shape[0]):
			K = int(self.col_partition[i].max()+1)
			self.N += int(pow(Bn,K))

		self.states = []
		state_idx = 0
		for col_part in range(self.col_partition.shape[0]):
			K = int(max(self.col_partition[col_part])+1)
			n_views = int(max(self.col_partition[col_part])+1)
			# self.states[state_idx]= dict()
			for rprt in range(len(self.row_perms[n_views-1])):
				temp_state = dict()
				temp_state['idx'] = state_idx
				temp_state['col_parts'] = self.col_partition[col_part]
				
				this_row_partition = self.row_perms[n_views-1][rprt][0]-1
				temp_row_parts = np.array([self.row_partition[this_row_partition]])
				for view in range(1,n_views):
					this_row_partition = self.row_perms[n_views-1][rprt][view]-1
					temp_row_parts = np.vstack((temp_row_parts,self.row_partition[this_row_partition]))
				temp_state['row_parts'] = temp_row_parts
				self.states.append(temp_state)
				
				state_idx += 1


	def getState(self, state_num):
		if state_num < 0 or state_num > self.N-1:
			return None
		else:
			return self.states[state_num]

	def findState(self, col_part, row_parts):

		n_views = len(row_parts)

		col_part_cm = vectorToCHMat(col_part);

		for state in range(len(self.states)):
			this_col_part = self.states[state]['col_parts']
			if max(this_col_part) != n_views-1:
				continue

			if not np.all(col_part_cm == vectorToCHMat(this_col_part)):
				continue

			for view in range(n_views):
				this_row_part = self.states[state]['row_parts'][view]
				if np.all(vectorToCHMat(row_parts[view])==vectorToCHMat(this_row_part)):
					if view == n_views-1:
						return self.states[state]['idx']
				else:
					break
			
		print("Error: no state match found")
		return None

	def test(self):
		print("Testing CrossCatPartitions")
		error = False
		# make sure findState returns the appropriate state number
		for state in self.states:
			cols = state['col_parts']
			rows = state['row_parts']
			found_index = self.findState(cols,rows)
			if state['idx'] != found_index:
				error = True
				print(" ")
				print("findState returned incorrect state (%i instead of %i). " % found_index, state['idx'])
				print("Found state: ")
				print("Cols")
				print(self.states[found_index]['col_parts'])
				print("Rows")
				print(self.states[found_index]['row_parts'])
				print("Actual state: ")
				print("Cols")
				print(cols)
				print("Rows")
				print(rows)
				print(" ")
			# make sure state mathces after set to zero index
			cols = cols - 1
			rows = rows - 1
			found_index = self.findState(cols,rows)
			if state['idx'] != found_index:
				error = True
				print(" ")
				print("findState returned incorrect for relabled state (%i instead of %i). " % found_index, state['idx'])
				print("Found state: ")
				print("Cols")
				print(self.states[found_index]['col_parts'])
				print("Rows")
				print(self.states[found_index]['row_parts'])
				print("Actual state: ")
				print("Cols")
				print(cols)
				print("Rows")
				print(rows)
				print(" ")

		if error:
			print("Test failed.")
		else:
			print("All tests passed.")

# partition class
# The thing we want is the EnumeratePartitions static  method. The Next function 
# doesn't actually stop itself so know that if you're going to try to use this 
# code elsewhere
class Partition(object):

	def __init__(self, N):
		self.N = N
		self.proceed = True
		self.s = np.ones(N, dtype=int)
		self.m = np.ones(N, dtype=int)
		
	# Enumerates the set of all partitionings of N points
	@staticmethod
	def EnumeratePartitions(N):
		p = Partition(N)

		expectedPartitions = Bell(N)
		currentPartition = 2

		C = np.copy(p.s-1)
	    
		while p.proceed:
			p.Next()
			if p.proceed:
				C = np.vstack([C,p.s-1]);
				currentPartition += 1
			else:
				break

			if currentPartition > expectedPartitions:
				break

		return C

	# generates the next partition
	def Next(self):
		n = self.N
		i = 0;
		self.s[i] = self.s[i] + 1;
		while (i < n) and (self.s[i] > self.m[i+1] + 1):
			self.s[i] = 1;
			i += 1;
			self.s[i] += 1
	    
		if self.s[i] > self.m[i]:
			self.m[i] = self.s[i]
        
		for j in range(i-1,-1,-1):
			self.m[j] = self.m[i]
        
		self.proceed = True;
	    

# returns B_N, the Nth Bell number. Uses the definition as a sum of Striling 
# numbers of the second kind. http://en.wikipedia.org/wiki/Bell_number
def Bell(N):
	B_N = 0.0;
	# range(n) produces an iterator over 0,1,...,n-1
	for k in range(N+1):
		snk = Stirling2nd(N,k)
		B_N += snk
	
	return B_N

# Returns the Striling number of the second kind. Math taken from Wikipedia:
# http://en.wikipedia.org/wiki/Stirling_numbers_of_the_second_kind
def Stirling2nd(n,k):
	snk = 0.0;
	const = (1.0/math.factorial(k));
	for j in range(k+1):
		p1 = math.pow(-1,k-j)
		p2 = sp.special.binom(k,j)
		p3 = math.pow(j,n)
		snk += p1*p2*p3
	
	return const*snk

# Log CRP
# log probability of the partitioning, prt, under the CRP with concentration 
# parameter, alpha.
def lcrp(prt,alpha):
	# generate a histogram of prt
	k = max(prt)+1
	ns = np.zeros(k)
	n = len(prt)
	for i in range(n):
		ns[prt[i]] += 1.0

	lp = sum(sp.special.gammaln(ns))+k*math.log(alpha)+sp.special.gammaln(alpha)-sp.special.gammaln(n+alpha)

	if np.any(np.isnan(lp)) or np.any(np.isinf(lp)):
		print("prt: ")
		print(prt)
		print("ns: ")
		print(ns)
		print("n: " + str(n))
		print("k: " + str(k))
		print(list(range(k)))
		print(" ")

	return lp

# Normal-Gamma marginal likelihood 
# Taken from Yee Whye Teh's "The Normal Exponential Family with 
# Normal-Inverse-Gamma Prior"
# http://www.stats.ox.ac.uk/~teh/research/notes/GaussianInverseGamma.pdf
def NGML(X,mu,r,nu,s):

	X = np.array(X.flatten(1))

	# constant
	LOGPI = np.log(math.pi)

	n = float(len(X)) 	# number of data points
	xbar = np.mean(X)	

	# update parameters
	rp  = r + n;
	nup = nu + n;
	mup = (r*mu+sum(X))/(r+n);
	spr = s + sum(X**2)+r*mu**2-rp*mup**2

	# the log answer
	lp1 = (-n/2)*LOGPI - (1/2)*np.log(rp)-(nup/2)*np.log(spr)+sp.special.gammaln(nup/2);
	lp2 = -(1/2)*np.log(r)-(nu/2)*np.log(s)+sp.special.gammaln(nu/2);
	lp = lp1-lp2;

	if np.isnan(lp):
		# print "X"
		# print(X)
		# print "X-xbar"
		# print(X-xbar)
		# print"sum((X-xbar)**2): %f" % sum((X-xbar)**2)
		print("xbar: %f" % xbar)
		print("nu_n: %f" % nu_n)
		print("mu_n: %f" % mu_n)
		print("k_n: %f" % k_n)
		print("s_n: %f" % s_n)
		print("lp1: %f" % lp1)
		print("lp2: %f" % lp2)

		sys.exit(0)


	return lp

# CrossCat Marginal Likelihood
# Computes the marginal likelihood of the array of data in ccmat given all 
# possible partitionings of columns and rows of ccmat into views and categories.
# Goes through each partitioning, divides the data up and sequentially sends 
# that data and the priors to NGML, then sums these answers (they're log).
# 	Takes the data array ccmat; the prior mean, M0; the prior variance, V0; the 
# inverse-gamma hyperparameters A0 and B0; and the CRP concentration parameter, 
# alpha.
# 	Returns the log marginal likelohood.
def CCML(ccpart,ccmat,mu,r,nu,s,row_alpha,col_alpha):
	lp = []
	
	ccmat = np.array(ccmat)

	state = ccpart.states[1]

	# loop through the states
	for state in ccpart.states:
		all_cols = state['col_parts']
		all_rows = state['row_parts']

		K = max(all_cols)+1
				
		lp_temp = lcrp(all_cols,col_alpha)
		for view in range(K):
			row_part = all_rows[view,:]
			lp_temp += lcrp(row_part,row_alpha)
			cols_view = np.nonzero(all_cols==view)[0]
			for col in cols_view:
				for cat in range(row_part.max()+1):
					X = ccmat[np.nonzero(row_part==cat)[0],col]
					lp_temp += NGML(X,mu,r,nu,s)

		lp.append(lp_temp);

	# return the normalized probabilities
	return lp-sp.misc.logsumexp(lp)

# Fixes the prior
def FixPriors(X_L,alpha,mu,s,r,nu):
	num_cols = len(X_L['column_partition']['assignments'])
	X_L['column_partition']['hypers']['alpha'] = alpha
	# print(new_X_L)
	for i in range(len(X_L['view_state'])):
		X_L['view_state'][i]['row_partition_model']['hypers']['alpha'] = alpha
	for i in range(num_cols):
		X_L['column_hypers'][i]['alpha'] = alpha
		X_L['column_hypers'][i]['mu']    = mu
		X_L['column_hypers'][i]['s']     = s
		X_L['column_hypers'][i]['r']     = r
		X_L['column_hypers'][i]['nu']    = nu

	return X_L

# Convert assignment vector to cohabitation matrix. A cohabitation matrix is an 
# N-by-N matrix where entry [i,j] = 1 is data points i and j belong to the same
# category and 0 otherwise. This function is used to match sampled states with
# enumerated states to compare the sampler with the enumerated answers.
def vectorToCHMat(col_partition):
	# print(col_partition)
	N = len(col_partition)
	
	chmat = np.zeros((N,N))
	for i in range(N):
		for j in range(N):
			if col_partition[i] == col_partition[j]:
				chmat[i,j] = 1
	return chmat



import random
import argparse
from functools import partial
#
import numpy
import pylab
#
from crosscat.LocalEngine import LocalEngine
import crosscat.tests.geweke_utils as gu
import crosscat.tests.plot_utils as pu
import crosscat.utils.data_utils as du
import crosscat.utils.convergence_test_utils as ctu
import experiment_runner.experiment_utils as eu


directory_prefix='test_log_likelihood'
#
noneify = set(['n_test'])
base_config = dict(
    gen_seed=0,
    num_rows=100, num_cols=4,
    num_clusters=5, num_views=1,
    n_steps=10, n_test=10,
    )

def runner(config):
    # helpers
    def munge_config(config):
        kwargs = config.copy()
        kwargs['num_splits'] = kwargs.pop('num_views')
        n_steps = kwargs.pop('n_steps')
        n_test = kwargs.pop('n_test')
        return kwargs, n_steps, n_test
    def calc_ll(T, p_State):
        log_likelihoods = map(p_State.calc_row_predictive_logp, T)
        mean_log_likelihood = numpy.mean(log_likelihoods)
        return mean_log_likelihood
    def gen_data(**kwargs):
        T, M_c, M_r, gen_X_L, gen_X_D = du.generate_clean_state(**kwargs)
        #
        engine = LocalEngine()
        sampled_T = gu.sample_T(engine, M_c, T, gen_X_L, gen_X_D)
        T_test = random.sample(sampled_T, n_test)
        gen_data_ll = ctu.calc_mean_test_log_likelihood(M_c, T, gen_X_L, gen_X_D, T)
        gen_test_set_ll = ctu.calc_mean_test_log_likelihood(M_c, T, gen_X_L, gen_X_D, T_test)
        #
        return T, M_c, M_r, T_test, gen_data_ll, gen_test_set_ll
    kwargs, n_steps, n_test = munge_config(config)
    T, M_c, M_r, T_test, gen_data_ll, gen_test_set_ll = gen_data(**kwargs)
    # set up to run inference
    calc_data_ll = partial(calc_ll, T)
    calc_test_set_ll = partial(calc_ll, T_test)
    diagnostic_func_dict = dict(
            data_ll=calc_data_ll,
            test_set_ll=calc_test_set_ll,
            )
    # run inference
    engine = LocalEngine()
    X_L, X_D = engine.initialize(M_c, M_r, T)
    X_L, X_D, diagnostics_dict = engine.analyze(M_c, T, X_L, X_D,
            do_diagnostics=diagnostic_func_dict, n_steps=n_steps)
    # package result
    final_data_ll = diagnostics_dict['data_ll'][-1][-1]
    final_test_set_ll = diagnostics_dict['test_set_ll'][-1][-1]
    summary = dict(
            gen_data_ll=gen_data_ll,
            gen_test_set_ll=gen_test_set_ll,
            final_data_ll=final_data_ll,
            final_test_set_ll=final_test_set_ll,
            )

    result = dict(
            config=config,
            summary=summary,
            diagnostics_dict=diagnostics_dict,
            )
    return result

def plotter(result):
    pylab.figure()
    diagnostics_dict = result['diagnostics_dict']
    gen_data_ll = result['summary']['gen_data_ll']
    gen_test_set_ll = result['summary']['gen_test_set_ll']
    #
    pylab.plot(diagnostics_dict['data_ll'], 'g')
    pylab.plot(diagnostics_dict['test_set_ll'], 'r')
    pylab.axhline(gen_data_ll, color='g', linestyle='--')
    pylab.axhline(gen_test_set_ll, color='r', linestyle='--')
    return

def _generate_parser():
    default_gen_seed = [0]
    default_num_rows = [100, 200, 500]
    default_num_cols = [10, 20]
    default_num_clusters = [1, 10, 20]
    default_num_views = [1, 5]
    default_n_steps = [10]
    default_n_test = [20]
    parser = argparse.ArgumentParser()
    parser.add_argument('--gen_seed', nargs='+', default=default_gen_seed, type=int)
    parser.add_argument('--num_rows', nargs='+', default=default_num_rows, type=int)
    parser.add_argument('--num_cols', nargs='+', default=default_num_cols, type=int)
    parser.add_argument('--num_clusters', nargs='+',
            default=default_num_clusters, type=int)
    parser.add_argument('--num_views', nargs='+', default=default_num_views,
            type=int)
    parser.add_argument('--n_steps', nargs='+', default=default_n_steps, type=int)
    parser.add_argument('--n_test', nargs='+', default=default_n_test, type=int)
    #
    parser.add_argument('--no_plots', action='store_true')
    parser.add_argument('--dirname', default='test_log_likelihood', type=str)
    return parser

def _munge_args(args):
    kwargs = args.__dict__.copy()
    do_plots = not kwargs.pop('no_plots')
    dirname = kwargs.pop('dirname')
    return kwargs, do_plots, dirname

def summary_plotter(results, dirname='./', save=True):
    def _scatter(x, y):
        pylab.figure()
        pylab.scatter(x, y)
        pylab.gca().set_aspect(1)
        xlim = pylab.gca().get_xlim()
        pylab.plot(xlim, xlim)
        return
    def _plot(frame, variable_suffix, filename=None, dirname='./'):
        x = frame['gen_' + variable_suffix]
        y = frame['final_' + variable_suffix]
        _scatter(x, y)
        pylab.title(variable_suffix)
        if filename is not None:
            pu.save_current_figure(filename, dir=dirname, close=True)
            pass
        return
    frame = eu.summaries_to_frame(results)
    for variable_suffix in ['test_set_ll', 'data_ll']:
        filename = variable_suffix if save else None
        _plot(frame, variable_suffix, filename=filename, dirname=dirname)
        pass
    return


if __name__ == '__main__':
    from experiment_runner.ExperimentRunner import ExperimentRunner, propagate_to_s3

    # parse args
    parser = _generate_parser()
    args = parser.parse_args()
    kwargs, do_plots, dirname = _munge_args(args)


    # create configs
    config_list = eu.gen_configs(base_config, **kwargs)


    # do experiments
    er = ExperimentRunner(runner, storage_type='fs',
            dirname_prefix=dirname,
            bucket_str='experiment_runner',
            )
    er.do_experiments(config_list)
    propagate_to_s3(er)


    if do_plots:
        results = er.get_results(er.frame).values()
        summary_plotter(results, dirname=dirname)
#        eu.plot_results(plotter, results, generate_dirname,
#                saver=pu.save_current_figure, filename='over_iters',
#                dirname=dirname)
